from autofit.core import model_mapper as mm
from autofit.core import non_linear as nl
from autocti.model import arctic_params as ap
from workspace.pipelines import parallel_x3

import matplotlib.pyplot as plt

def setup_multinest_results(model_mapper, name, input_model, sigma_limit=2.0, requirements=None):

    multinest = nl.MultiNest(model_mapper=model_mapper, name=name, sigma_limit=sigma_limit)

    most_likely = multinest.most_likely_from_summary()
    most_probable = multinest.most_probable_from_summary()
    upper_limits = multinest.model_at_upper_sigma_limit(sigma_limit=sigma_limit)
    lower_limits = multinest.model_at_lower_sigma_limit(sigma_limit=sigma_limit)

    upper_error = list(map(lambda mp, upper : upper - mp, most_probable, upper_limits))
    lower_error = list(map(lambda mp, lower : mp - lower, most_probable, lower_limits))
    errors = multinest.model_errors_at_sigma_limit(sigma_limit=sigma_limit)
    error_precision = list(map(lambda upper, lower : upper - lower, upper_limits, lower_limits))

    most_probable_difference = list(map(lambda input, mp :  mp - input, input_model, most_probable))
    upper_limit_difference = list(map(lambda input, upper : upper - input, input_model, upper_limits))
    lower_limit_difference = list(map(lambda input, lower : lower - input, input_model, lower_limits))

    # print('Model Accuracy')
    #
    # for i in range(multinest.variable.prior_count):
    #    print()
    #    print('input = ' + str(input_model[i]) + '  most probable = ' + str(most_probable[i]) +
    #          ' (' + str(lower_limits[i]) + ', ' + str(upper_limits[i]) + ')')
    #    print('error precision = ' + str(error_precision[i]))
    #    print(str(lower_limit_difference[i]) + ' ' + str(most_probable_difference[i]) + ' ' + str(upper_limit_difference[i]))

    #    if lower_limits[i] > most_probable[i]:
    #        stop
    #
    # print()
    # print('Model Requirements')

    # print()
    # print(list(map(lambda requirement, error : error / requirement, requirements, errors)))

    return most_likely, most_probable, lower_limits, upper_limits, upper_error, lower_error, error_precision


from autofit.core import model_mapper as mm
from autocti.model import arctic_params as ap
import workspace_jam.scripts.requirements as req

import matplotlib.pyplot as plt

model_mapper = mm.ModelMapper(parallel=ap.ParallelTwoSpecies)
model_mapper.parallel.well_fill_alpha = 0.0
model_mapper.parallel.well_fill_gamma = 0.0

input_model = [10000.0, 0.58, 0.13, 0.25, 1.25, 4.4]
sigma_limit = 2.0

most_likely_lr, most_probable_lr, lower_limits_lr, upper_limits_lr, lower_error_lr, upper_error_lr, error_pecision_lr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Parallel_lr_pn/')

most_likely_mr, most_probable_mr, lower_limits_mr, upper_limits_mr, lower_error_mr, upper_error_mr, error_pecision_mr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Parallel_mr_pn/')

most_likely_hr, most_probable_hr, lower_limits_hr, upper_limits_hr, lower_error_hr, upper_error_hr, error_pecision_hr = \
req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model,
                                                sigma_limit=sigma_limit, name='Results/Parallel_hr_pn/')

most_likely_x2,  most_probable_x2, lower_limits_x2, upper_limits_x2, lower_error_x2, upper_error_x2,  error_pecision_x2 = \
req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model,
                                                sigma_limit=sigma_limit, name='Results/Parallel_x2_pn/')

ylabels = ['d', r'$\beta$', r'$\rho_{1}$', r'$\rho_{2}$', r'$\tau_{1}$', r'$\tau_{2}$']

plt.figure(figsize=(20, 15))
plt.suptitle('Accuracy of Parallel CTI Model', fontsize=20)

for i in range(6):
    plt.subplot(2, 3, i+1)
    plt.plot([517, 1034, 2068, 4136], [input_model[i], input_model[i], input_model[i], input_model[i]], linestyle='--')
    plt.errorbar(x=[517, 1034, 2068, 4136], y=[most_probable_lr[i], most_probable_mr[i],
                                               most_probable_hr[i], most_probable_x2[i]],
                 yerr=[[lower_error_lr[i], lower_error_mr[i], lower_error_hr[i], lower_error_x2[i]],
                       [upper_error_lr[i], upper_error_mr[i], upper_error_hr[i], upper_error_x2[i]]])
    plt.xlabel('Number of columns', fontsize=12)
    plt.ylabel(ylabels[i], fontsize=12)

plt.show()

plt.figure(figsize=(20, 15))
plt.suptitle('Precision of Parallel CTI Model', fontsize=20)

for i in range(6):
    plt.subplot(2, 3, i+1)
    plt.plot([517, 1034, 2068, 4136], [error_pecision_lr[i], error_pecision_mr[i],
                                       error_pecision_hr[i], error_pecision_x2[i]])
    plt.xlabel('Number of columns', fontsize=12)
    plt.ylabel(ylabels[i], fontsize=12)

plt.show()


#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/util_test.py

Created on: 04/23/18
Author: user
"""

import os
import shutil

import numpy as np
import pytest

from autocti.data import util

path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))

test_data_path = "{}/../test_files/array/".format(os.path.dirname(os.path.realpath(__file__)))


class TestFits:

    def test__numpy_array_from_fits__3x3_all_ones(self):
        arr = util.numpy_array_2d_from_fits(file_path=test_data_path + '3x3_ones.fits', hdu=0)

        assert (arr == np.ones((3, 3))).all()

    def test__numpy_array_from_fits__4x3_all_ones(self):
        arr = util.numpy_array_2d_from_fits(file_path=test_data_path + '4x3_ones.fits', hdu=0)

        assert (arr == np.ones((4, 3))).all()

    def test__numpy_array_to_fits__output_and_load(self):
        if os.path.exists(test_data_path + 'test.fits'):
            os.remove(test_data_path + 'test.fits')

        arr = np.array([[10., 30., 40.],
                        [92., 19., 20.]])

        util.numpy_array_2d_to_fits(arr, file_path=test_data_path + 'test.fits')

        array_load = util.numpy_array_2d_from_fits(file_path=test_data_path + 'test.fits', hdu=0)

        assert arr == pytest.approx(array_load, 1.0e-4)



import workspace_jam.scripts.requirements.requirement_converter as convert

import numpy as np

input_model = [0.01, 0.03, 0.9, 0.8, 3.5, 20.0, 1.0e-4, 0.58]
input_rho = np.array([0.01, 0.03, 0.9])
input_tau = np.array([0.8, 3.5, 20.0])

most_probable = np.array([0.024758, 0.88427, 1.28204, 18.6937, 3.5, 9.58])
upper_limits = np.array([0.04782, 0.9442, 2.5276, 21.8204, 29.4351, 0.5980])
lower_limits = np.array([0.0144, 0.84705, 0.69958, 17.2769, -31.104, 0.55735])

sigma_limit = 2.0

delta_ellipticity_true = convert.convert_to_ellipticity(rho=input_rho, tau=input_tau)

model_rho = np.array([most_probable[0], most_probable[1]])
model_tau = np.array([most_probable[2], most_probable[3]])
delta_ellipticity_model = convert.convert_to_ellipticity(rho=model_rho, tau=model_tau)

upper_rho = np.array([upper_limits[0], upper_limits[1]])
upper_tau = np.array([upper_limits[2], upper_limits[3]])
delta_ellipticity_upper = convert.convert_to_ellipticity(rho=upper_rho, tau=upper_tau)

lower_rho = np.array([lower_limits[0], lower_limits[1]])
lower_tau = np.array([lower_limits[2], lower_limits[3]])
delta_ellipticity_lower = convert.convert_to_ellipticity(rho=lower_rho, tau=lower_tau)

print('Model True')
print(delta_ellipticity_true)
print(np.sum(delta_ellipticity_true))
print()

print('Model most probable')
print(delta_ellipticity_model)
print(np.sum(delta_ellipticity_model))
print()

print('Model upper')
print(delta_ellipticity_upper)
print(np.sum(delta_ellipticity_upper))
print()

print('Model lower')
print(delta_ellipticity_lower)
print(np.sum(delta_ellipticity_lower))
print()

print('Induced Delta Elliptcitiy')
print(np.abs(np.sum(delta_ellipticity_model) - np.sum(delta_ellipticity_true)))
print(np.abs(np.sum(delta_ellipticity_upper) - np.sum(delta_ellipticity_true)))
print(np.abs(np.sum(delta_ellipticity_lower) - np.sum(delta_ellipticity_true)))
print()

print('Induced Delta Elliptcitiy / Requirement')
print(np.abs(np.sum(delta_ellipticity_model) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print(np.abs(np.sum(delta_ellipticity_upper) - np.sum(delta_ellipticity_true)) / (1.1e-4))
print(np.abs(np.sum(delta_ellipticity_lower) - np.sum(delta_ellipticity_true)) / (1.1e-4))


#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/VIS_CTI_ChargeInjection/CIImage.pyy

Created on: 02/14/18
Author: James Nightingale
"""

import numpy as np

from autocti.charge_injection import ci_frame


class Mask(np.ndarray):

    def __new__(cls, array, *args, **kwargs):
        mask = np.array(array, dtype='float64').view(cls)
        return mask

    @classmethod
    def empty_for_shape(cls, shape, **kwargs):
        """
        Create the mask used for CTI Calibration as all False's (e.g. no masking).

        Parameters
        ----------
        shape : (int, int)
            The dimensions of the 2D mask.
        """
        # noinspection PyArgumentList
        return cls(array=np.full(shape=shape, fill_value=False))

    @classmethod
    def from_masked_regions(cls, shape, masked_regions, **kwargs):

        mask = cls.empty_for_shape(shape)
        masked_regions = list(map(lambda r: ci_frame.Region(r), masked_regions))
        for region in masked_regions:
            mask[region.y0:region.y1, region.x0:region.x1] = True

        return mask

    @classmethod
    def from_cosmic_ray_image(cls, shape, frame_geometry, cosmic_ray_image, cosmic_ray_parallel_buffer=0,
                              cosmic_ray_serial_buffer=0, cosmic_ray_diagonal_buffer=0, **kwargs):
        """
        Create the mask used for CTI Calibration, which is all False unless specific regions are input for masking.

        Parameters
        ----------
        shape : (int, int)
            The dimensions of the 2D mask.
        frame_geometry : ci_frame.CIQuadGeometry
            The quadrant geometry of the simulated image, defining where the parallel / serial overscans are and \
            therefore the direction of clocking and rotations before input into the cti algorithm.
        cosmic_ray_image : ndarray
            2D array flagging where cosmic rays on the image.
        cosmic_ray_parallel_buffer : int
            If a cosmic-ray mask is supplied, the number of pixels from each ray pixels are masked in the parallel \
            direction.
        cosmic_ray_serial_buffer : int
            If a cosmic-ray mask is supplied, the number of pixels from each ray pixels are masked in the serial \
            direction.
        """
        mask = cls.empty_for_shape(shape)

        cosmic_ray_mask = (cosmic_ray_image > 0.0).astype('bool')

        for y in range(mask.shape[0]):
            for x in range(mask.shape[1]):
                if cosmic_ray_mask[y, x]:
                    y0, y1 = frame_geometry.parallel_trail_from_y(y, cosmic_ray_parallel_buffer)
                    mask[y0:y1, x] = True
                    x0, x1 = frame_geometry.serial_trail_from_x(x, cosmic_ray_serial_buffer)
                    mask[y, x0:x1] = True
                    y0, y1 = frame_geometry.parallel_trail_from_y(y, cosmic_ray_diagonal_buffer)
                    x0, x1 = frame_geometry.serial_trail_from_x(x, cosmic_ray_diagonal_buffer)
                    mask[y0:y1, x0:x1] = True

        return mask



from autofit import conf
import os

# Setup the path to the workspace, using a relative directory name.
workspace_path = '{}/../'.format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
conf.instance = conf.Config(config_path=workspace_path + 'config', output_path=workspace_path + 'output')

from autofit.tools import path_util
from autocti.data import mask as msk
from autocti.charge_injection import ci_frame
from autocti.charge_injection import ci_pattern
from autocti.charge_injection import ci_data

from autocti.model import arctic_settings

from multiprocessing import Pool

# Welcome to the pipeline runner. This tool allows you to CTI calibration data on strong lenses, and pass it to
# pipelines for a PyAutoCTI analysis. To show you around, we'll load up some example data and run it through some of
# the example pipelines that come distributed with PyAutoCTI.

# The runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out data-set names
# and pipelines to perform different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# The pipeline runner is fairly self explanatory. Make sure to checkout the pipelines in the
#  workspace/pipelines/examples/ folder - they come with detailed descriptions of what they do. I hope that you'll
# expand on them for your own personal scientific needs

# It is convenient to specify the image type and model used to simulate that image as strings, so that if the
# pipeline is applied to multiple images we don't have to change all of the path entries in the
# load_ci_data_from_fits function below.

ci_data_type = 'ci_images_uniform' # Charge injection data consisting of 2 images with uniform injections.
ci_data_model = 'parallel_x2_species' # Shows the data was creating using a parallel CTI model with one species.
ci_data_resolution = 'low_res' # The resolution of the image.

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/ci_images_uniform/serial_x3_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

if ci_data_resolution in 'high_res':
    shape = (2316, 2119)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))
elif ci_data_resolution in 'mid_res':
    shape = (2316, 1034)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 1014)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 1014, 1034)))
elif ci_data_resolution in 'low_res':
    shape = (2316, 517)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 497)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 497, 517)))

# The charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, 30, 51, shape[1]-20), (330, 360, 51, shape[1]-20),
              (660, 690, 51, shape[1]-20), (990, 1020, 51, shape[1]-20),
              (1320, 1350, 51, shape[1]-20), (1650, 1680, 51, shape[1]-20),
              (1980, 2010, 51, shape[1]-20)]

# The normalization of the charge injection for each image.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# Create the charge injection pattern objects used for this pipeline.
patterns = ci_pattern.uniform_from_lists(normalizations=normalizations, regions=ci_regions)

# There are 8 images to load, which is what the normalizations list tells us. To load them, its easiest to create a
# 'data' list and then iterate over a for loop, to append each set of data to the list of data we pass to the pipeline
# when we run it.

datas = []

for index, normalization in enumerate(normalizations):

    datas.append(ci_data.ci_data_from_fits(
                 frame_geometry=frame_geometry, ci_pattern=patterns[index],
                 image_path=ci_data_path + 'image_' + str(int(normalization)) + '.fits',
                 ci_pre_cti_path=ci_data_path +'/ci_pre_cti_' + str(int(normalization)) + '.fits',
                 noise_map_from_single_value=4.0))

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For parallel ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
parallel_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=True, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(parallel=parallel_cti_settings)

def mask_function_for_bug(shape, frame_geometry):
    return msk.Mask.from_masked_regions(shape=shape, frame_geometry=frame_geometry,
                                        masked_regions=[(0, 330, 0, shape[1])])

# Running a pipeline is easy, we simply import it from the pipelines folder and pass the ci data to its run function.
# Below, we'll' use a 2 phase example pipeline to fit the data with a one species parallel CTI model.
# Checkout workspace/pipelines/examples/parallel_x1_species.py' for a full description of the pipeline.

# The pool command tells our code to parallelize the analysis over 2 CPU's, where each CPU fits a different charge
# injection image

from workspace.pipelines import parallel_x2_species
pipeline = parallel_x2_species.make_pipeline(phase_folders=[ci_data_type, ci_data_model, ci_data_resolution],
                                             mask_function=mask_function_for_bug)
pipeline.run(ci_datas=datas, cti_settings=cti_settings, pool=Pool(processes=2))


import numpy as np
from autocti.charge_injection.ci_frame import Region


class MockPattern(object):

    def __init__(self, regions=None):

        self.regions = regions


class MockGeometry(object):

    def __init__(self):
        super(MockGeometry, self).__init__()


class MockFrameGeometry(object):

    def __init__(self, value=1.0):
        self.value = value

    def add_cti(self, image, cti_params, cti_settings):
        return self.value * np.ones((2, 2))

class MockCIFrame(object):

    def __init__(self, value=1.0):

        self.frame_geometry = MockFrameGeometry(value=value)
        self.value = value

    def ci_regions_from_array(self, array):
        return array[0:2, 0]

    def parallel_non_ci_regions_frame_from_frame(self, array):
        return array[0:2, 1]

    def serial_all_trails_frame_from_frame(self, array):
        return array[0, 0:2]

    def serial_overscan_above_trails_frame_from_frame(self, array):
        return array[1, 0:2]


class MockCIGeometry(object):

    def __init__(self, serial_prescan=(0, 1, 0, 1), serial_overscan=(0, 1, 0, 1)):
        super(MockCIGeometry, self).__init__()
        self.serial_prescan = Region(serial_prescan)
        self.serial_overscan = Region(serial_overscan)


class MockCIPreCTI(np.ndarray):

    def __new__(cls, array, frame_geometry=MockGeometry(), ci_pattern=MockPattern(), value=1.0, *args, **kwargs):
        ci = np.array(array).view(cls)
        ci.frame_geometry = frame_geometry
        ci.ci_pattern = ci_pattern
        ci.value = value
        return ci

    def ci_post_cti_from_cti_params_and_settings(self, cti_params, cti_settings):
        return self.value * np.ones((2, 2))


class MockParams(object):

    def __init__(self):
        pass


class MockSettings(object):

    def __init__(self):
        pass


class MockChInj(np.ndarray):

    def __new__(cls, array, geometry=None, ci_pattern=None, *args, **kwargs):
        ci = np.array(array).view(cls)
        ci.frame_geometry = geometry
        ci.ci_pattern = ci_pattern
        return ci


from functools import partial

import numpy as np
from autofit import conf
from autofit.mapper import prior_model as pm
from autofit.optimize import non_linear as nl
from autofit.tools import phase as ph
from autofit.tools import phase_property

from autocti import exc
from autocti.charge_injection import ci_fit, ci_data, ci_hyper
from autocti.charge_injection.plotters import ci_data_plotters
from autocti.charge_injection.plotters import ci_fit_plotters
from autocti.data import mask as msk
from autocti.model import arctic_params


def cti_params_for_instance(instance):
    return arctic_params.ArcticParams(
        parallel_ccd=instance.parallel_ccd if hasattr(instance, "parallel_ccd") else None,
        parallel_species=instance.parallel_species if hasattr(instance, "parallel_species") else None,
        serial_ccd=instance.serial_ccd if hasattr(instance, "serial_ccd") else None,
        serial_species=instance.serial_species if hasattr(instance, "serial_species") else None)


class ConsecutivePool(object):
    """
    Replicates the interface of a multithread pool but performs computations consecutively
    """

    @staticmethod
    def map(func, ls):
        return map(func, ls)


class Phase(ph.AbstractPhase):

    def make_result(self, result, analysis):
        return self.__class__.Result(constant=result.constant, figure_of_merit=result.figure_of_merit,
                                     variable=result.variable, analysis=analysis, optimizer=self.optimizer)

    def __init__(self, phase_name, phase_tag=None, phase_folders=None, optimizer_class=nl.DownhillSimplex,
                 mask_function=msk.Mask.empty_for_shape, columns=None, rows=None,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase in an analysis pipeline. Uses the set NonLinear optimizer to try to fit models and images passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a NonLinear optimizer
            The side length of the subgrid
        """
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         optimizer_class=optimizer_class)
        self.mask_function = mask_function
        self.columns = columns
        self.rows = rows
        self.cosmic_ray_parallel_buffer = cosmic_ray_parallel_buffer
        self.cosmic_ray_serial_buffer = cosmic_ray_serial_buffer
        self.cosmic_ray_diagonal_buffer = cosmic_ray_diagonal_buffer

    @property
    def constant(self):
        """
        Convenience method

        Returns
        -------
        ModelInstance
            A model instance comprising all the constant objects in this analysis
        """
        return self.optimizer.constant

    @property
    def variable(self):
        """
        Convenience method

        Returns
        -------
        ModelMapper
            A model mapper comprising all the variable (prior) objects in this analysis
        """
        return self.optimizer.variable

    def run(self, ci_datas, cti_settings, results=None, pool=None):
        """
        Run this phase.

        Parameters
        ----------
        pool
        cti_settings
        ci_datas
        results: [Result]
            An object describing the results of the last phase or None if no phase has been executed

        Returns
        -------
        result: Phase.Result
            A result object comprising the best fit model and other ci_data.
        """

        analysis = self.make_analysis(ci_datas=ci_datas, cti_settings=cti_settings, results=results,
                                      pool=pool)

        result = self.optimizer.fit(analysis)

        return self.make_result(result=result, analysis=analysis)

    # noinspection PyMethodMayBeStatic
    def extract_ci_data(self, data, mask):
        return ci_data.MaskedCIData(image=data.image, noise_map=data.noise_map, ci_pre_cti=data.ci_pre_cti, mask=mask,
                                    ci_pattern=data.ci_pattern, ci_frame=data.ci_frame)

    def make_analysis(self, ci_datas, cti_settings, results=None, pool=None):
        """
        Create an analysis object. Also calls the prior passing and image modifying functions to allow child classes to
        change the behaviour of the phase.

        Parameters
        ----------
        cti_settings
        ci_datas
        pool
        results: [Results]
            The result from the previous phase

        Returns
        -------
        analysis: Analysis
            An analysis object that the non-linear optimizer calls to determine the fit of a set of values
        """
        masks = list(map(lambda data: self.mask_function(shape=data.image.shape,
                                                         frame_geometry=data.ci_frame.frame_geometry), ci_datas))

        cosmic_ray_masks = list(map(lambda data : msk.Mask.from_cosmic_ray_image(
            shape=data.shape, frame_geometry=data.ci_frame.frame_geometry,
            cosmic_ray_image=data.cosmic_ray_image,
            cosmic_ray_parallel_buffer=self.cosmic_ray_parallel_buffer,
            cosmic_ray_serial_buffer=self.cosmic_ray_serial_buffer,
            cosmic_ray_diagonal_buffer=self.cosmic_ray_diagonal_buffer)
        if data.cosmic_ray_image is not None else None, ci_datas))

        masks = list(map(lambda mask, cosmic_ray_mask :
                         mask + cosmic_ray_mask if cosmic_ray_mask is not None else mask,
                         masks, cosmic_ray_masks))

        ci_datas_fit = [self.extract_ci_data(data=data, mask=mask) for data, mask in zip(ci_datas, masks)]
        ci_datas_full = list(map(lambda data, mask:
                                 ci_data.MaskedCIData(image=data.image, noise_map=data.noise_map,
                                                      ci_pre_cti=data.ci_pre_cti, mask=mask,
                                                      ci_pattern=data.ci_pattern, ci_frame=data.ci_frame),
                                 ci_datas, masks))

        self.pass_priors(results)
        analysis = self.__class__.Analysis(ci_datas_extracted=ci_datas_fit, ci_datas_full=ci_datas_full,
                                           cti_settings=cti_settings, phase_name=self.phase_name,
                                           results=results, pool=pool)
        return analysis

    # noinspection PyAbstractClass
    class Analysis(nl.Analysis):

        def __init__(self, ci_datas_extracted, ci_datas_full, cti_settings, phase_name, results=None,
                     pool=None):
            """
            An analysis object. Once set up with the image ci_data (image, mask, noises) and pre-cti image it takes a \
            set of objects describing a model and determines how well they fit the image.

            Params
            ----------
            ci_data : [CIImage.CIImage]
                The charge injection ci_data-sets.
            """

            self.ci_datas_extracted = ci_datas_extracted
            self.ci_datas_full = ci_datas_full
            self.cti_settings = cti_settings
            self.phase_name = phase_name
            self.phase_output_path = "{}/{}".format(conf.instance.output_path, self.phase_name)

            self.pool = pool or ConsecutivePool
            self.results = results

            self.plot_count = 0

            def output_bool(name):
                return conf.instance.general.get('output', name, bool)

            self.extract_array_from_mask = output_bool('extract_images_from_mask')

            self.plot_ci_data_as_subplot = output_bool('plot_ci_data_as_subplot')
            self.plot_ci_data_image = output_bool('plot_ci_data_image')
            self.plot_ci_data_noise_map = output_bool('plot_ci_data_noise_map')
            self.plot_ci_data_ci_pre_cti = output_bool('plot_ci_data_ci_pre_cti')
            self.plot_ci_data_signal_to_noise_map = output_bool('plot_ci_data_signal_to_noise_map')

            self.plot_ci_fit_all_at_end_png = output_bool('plot_ci_fit_all_at_end_png')
            self.plot_ci_fit_all_at_end_fits = output_bool('plot_ci_fit_all_at_end_fits')

            self.plot_ci_fit_as_subplot = output_bool('plot_ci_fit_as_subplot')
            self.plot_ci_fit_image = output_bool('plot_ci_fit_image')
            self.plot_ci_fit_noise_map = output_bool('plot_ci_fit_noise_map')
            self.plot_ci_fit_signal_to_noise_map = output_bool('plot_ci_fit_signal_to_noise_map')
            self.plot_ci_fit_ci_pre_cti = output_bool('plot_ci_fit_ci_pre_cti')
            self.plot_ci_fit_ci_post_cti = output_bool('plot_ci_fit_ci_post_cti')
            self.plot_ci_fit_residual_map = output_bool('plot_ci_fit_residual_map')
            self.plot_ci_fit_chi_squared_map = output_bool('plot_ci_fit_chi_squared_map')

        @property
        def last_results(self):
            if self.results is not None:
                return self.results.last

        def visualize(self, instance, image_path, during_analysis):

            self.plot_count += 1

            ci_data_index = 0

            if self.plot_ci_data_as_subplot:

                ci_data_plotters.plot_ci_subplot(
                    ci_data=self.ci_datas_extracted[ci_data_index],
                    extract_array_from_mask=self.extract_array_from_mask,
                    output_path=image_path, output_format='png')

            ci_data_plotters.plot_ci_data_individual(
                ci_data=self.ci_datas_extracted[ci_data_index], extract_array_from_mask=self.extract_array_from_mask,
                should_plot_image=self.plot_ci_data_image,
                should_plot_noise_map=self.plot_ci_data_noise_map,
                should_plot_ci_pre_cti=self.plot_ci_data_ci_pre_cti,
                should_plot_signal_to_noise_map=self.plot_ci_data_signal_to_noise_map,
                output_path=image_path, output_format='png')

            fits = self.fits_of_ci_data_extracted_for_instance(instance=instance)

            ci_fit_index = 0

            if self.plot_ci_fit_as_subplot:
                ci_fit_plotters.plot_fit_subplot(
                    fit=fits[ci_fit_index], extract_array_from_mask=self.extract_array_from_mask,
                    output_path=image_path, output_format='png')

            if during_analysis:
                ci_fit_plotters.plot_fit_individuals(
                    fit=fits[ci_fit_index],
                    extract_array_from_mask=self.extract_array_from_mask,
                    should_plot_image=self.plot_ci_fit_image,
                    should_plot_noise_map=self.plot_ci_fit_noise_map,
                    should_plot_signal_to_noise_map=self.plot_ci_fit_signal_to_noise_map,
                    should_plot_ci_pre_cti=self.plot_ci_data_ci_pre_cti,
                    should_plot_ci_post_cti=self.plot_ci_fit_ci_post_cti,
                    should_plot_residual_map=self.plot_ci_fit_residual_map,
                    should_plot_chi_squared_map=self.plot_ci_fit_chi_squared_map,
                    output_path=image_path, output_format='png')

            elif not during_analysis:

                if self.plot_ci_fit_all_at_end_png:
                    ci_fit_plotters.plot_fit_individuals(
                        fit=fits[ci_fit_index],
                        extract_array_from_mask=self.extract_array_from_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_signal_to_noise_map=True,
                        should_plot_ci_pre_cti=True,
                        should_plot_ci_post_cti=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        output_path=image_path, output_format='png')

                if self.plot_ci_fit_all_at_end_fits:
                    ci_fit_plotters.plot_fit_individuals(
                        fit=fits[ci_fit_index],
                        extract_array_from_mask=self.extract_array_from_mask,
                        should_plot_image=True,
                        should_plot_noise_map=True,
                        should_plot_signal_to_noise_map=True,
                        should_plot_ci_pre_cti=True,
                        should_plot_ci_post_cti=True,
                        should_plot_residual_map=True,
                        should_plot_chi_squared_map=True,
                        output_path="{}/fits/".format(image_path), output_format='fits')

            return fits

        def fit(self, instance):
            """
            Determine the fitness of a particular model

            Parameters
            ----------
            instance

            Returns
            -------
            fit: ci_fit.Fit
                How fit the model is and the model
            """
            cti_params = cti_params_for_instance(instance=instance)
            self.check_trap_lifetimes_are_ascending(cti_params=cti_params)
            pipe_cti_pass = partial(pipe_cti, cti_params=cti_params, cti_settings=self.cti_settings)
            likelihood = np.sum(list(self.pool.map(pipe_cti_pass, self.ci_datas_extracted)))
            return likelihood

        def check_trap_lifetimes_are_ascending(self, cti_params):

            if hasattr(cti_params, 'serial'):

                if len(cti_params.serial) == 2:

                    if cti_params.serial_species[0].trap_lifetime > cti_params.serial_species[1].trap_lifetime:
                        raise exc.PriorException

                elif len(cti_params.serial) == 3:

                    if cti_params.serial_species[0].trap_lifetime > cti_params.serial_species[1].trap_lifetime or \
                            cti_params.serial_species[0].trap_lifetime > cti_params.serial_species[2].trap_lifetime or \
                            cti_params.serial_species[1].trap_lifetime > cti_params.serial_species[2].trap_lifetime:
                        raise exc.PriorException

        def fits_of_ci_data_extracted_for_instance(self, instance):
            cti_params = cti_params_for_instance(instance=instance)
            return list(map(lambda ci_data_fit:
                            ci_fit.CIFit(
                                masked_ci_data=ci_data_fit, cti_params=cti_params, cti_settings=self.cti_settings),
                            self.ci_datas_extracted))

        def fits_of_ci_data_full_for_instance(self, instance):
            cti_params = cti_params_for_instance(instance=instance)
            return list(map(lambda data:
                            ci_fit.CIFit(
                                masked_ci_data=data, cti_params=cti_params, cti_settings=self.cti_settings),
                            self.ci_datas_full))

    class Result(nl.Result):

        # noinspection PyUnusedLocal
        def __init__(self, constant, figure_of_merit, variable, analysis, optimizer):
            """
            The result of a phase
            """

            super(Phase.Result, self).__init__(constant=constant, figure_of_merit=figure_of_merit, variable=variable)

            self.analysis = analysis
            self.optimizer = optimizer

        @property
        def most_likely_extracted_fits(self):
            return self.analysis.fits_of_ci_data_extracted_for_instance(instance=self.constant)

        @property
        def most_likely_full_fits(self):
            return self.analysis.fits_of_ci_data_full_for_instance(instance=self.constant)

        @property
        def noise_scaling_maps_of_ci_regions(self):
            return list(map(lambda most_likely_full_fit:
                            most_likely_full_fit.noise_scaling_map_of_ci_regions,
                            self.most_likely_full_fits))

        @property
        def noise_scaling_maps_of_parallel_trails(self):
            return list(map(lambda most_likely_full_fit:
                            most_likely_full_fit.noise_scaling_map_of_parallel_trails,
                            self.most_likely_full_fits))

        @property
        def noise_scaling_maps_of_serial_trails(self):
            return list(map(lambda most_likely_full_fit:
                            most_likely_full_fit.noise_scaling_map_of_serial_trails,
                            self.most_likely_full_fits))

        @property
        def noise_scaling_maps_of_serial_overscan_above_trails(self):
            return list(map(lambda most_likely_full_fit:
                            most_likely_full_fit.noise_scaling_map_of_serial_overscan_above_trails,
                            self.most_likely_full_fits))


class ParallelPhase(Phase):

    parallel_species = phase_property.PhasePropertyCollection("parallel_species")
    parallel_ccd = phase_property.PhaseProperty("parallel_ccd")

    def __init__(self, phase_name, phase_tag=None, phase_folders=None, parallel_species=(), parallel_ccd=None, optimizer_class=nl.MultiNest,
                 mask_function=msk.Mask.empty_for_shape, columns=None,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase with a simple source/CTI model
        """
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         optimizer_class=optimizer_class, mask_function=mask_function, columns=columns, rows=None,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)
        self.parallel_species = parallel_species
        self.parallel_ccd = parallel_ccd

    def extract_ci_data(self, data, mask):
        return data.parallel_calibration_data(
            columns=(0, self.columns or data.ci_frame.frame_geometry.parallel_overscan.total_columns),
            mask=mask)

    class Analysis(Phase.Analysis):

        def check_trap_lifetimes_are_ascending(self, cti_params):
            trap_lifetimes = [parallel_species.trap_lifetime for parallel_species in cti_params.parallel_species]

            if not sorted(trap_lifetimes) == trap_lifetimes:
                raise exc.PriorException

        def describe(self, instance):
            return ("\nRunning CTI analysis for... \n\n"
                    "Parallel CTI: \n"
                    "Parallel Species:\n{}\n\n"
                    "Parallel CCD\n{}\n\n".format(instance.parallel_species, instance.parallel_ccd))


class SerialPhase(Phase):
    serial_species = phase_property.PhasePropertyCollection("serial_species")
    serial_ccd = phase_property.PhaseProperty("serial_ccd")

    def __init__(self, phase_name, phase_tag=None, phase_folders=None, serial_species=(), serial_ccd=None, optimizer_class=nl.MultiNest,
                 mask_function=msk.Mask.empty_for_shape, rows=None,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase with a simple source/CTI model
        """
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         optimizer_class=optimizer_class, mask_function=mask_function, rows=rows,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)
        self.serial_species = serial_species
        self.serial_ccd = serial_ccd

    def extract_ci_data(self, data, mask):
        return data.serial_calibration_data(rows=self.rows or (0, data.ci_pattern.regions[0].total_rows),
                                            mask=mask)

    class Analysis(Phase.Analysis):

        def check_trap_lifetimes_are_ascending(self, cti_params):
            trap_lifetimes = [serial_species.trap_lifetime for serial_species in cti_params.serial_species]

            if not sorted(trap_lifetimes) == trap_lifetimes:
                raise exc.PriorException

        def describe(self, instance):
            return (
                "\nRunning CTI analysis for... \n\n"
                "Serial CTI: \n"
                "Serial Species:\n{}\n\n"
                "Serial CCD\n{}\n\n".format(instance.serial_species, instance.serial_ccd))


class ParallelSerialPhase(Phase):
    parallel_species = phase_property.PhasePropertyCollection("parallel_species")
    serial_species = phase_property.PhasePropertyCollection("serial_species")
    parallel_ccd = phase_property.PhaseProperty("parallel_ccd")
    serial_ccd = phase_property.PhaseProperty("serial_ccd")

    def __init__(self, phase_name, phase_tag=None, phase_folders=None, parallel_species=(), serial_species=(),
                 parallel_ccd=None, serial_ccd=None,
                 optimizer_class=nl.MultiNest, mask_function=msk.Mask.empty_for_shape,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase with a simple source/CTI model

        Parameters
        ----------
        optimizer_class: class
            The class of a non-linear optimizer
        """
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         optimizer_class=optimizer_class, mask_function=mask_function, columns=None, rows=None,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)

        self.parallel_species = parallel_species
        self.serial_species = serial_species
        self.parallel_ccd = parallel_ccd
        self.serial_ccd = serial_ccd

    def extract_ci_data(self, data, mask):
        return data.parallel_serial_calibration_data(mask)

    class Analysis(Phase.Analysis):

        def check_trap_lifetimes_are_ascending(self, cti_params):

            trap_lifetimes = [parallel_species.trap_lifetime for parallel_species in cti_params.parallel_species]

            if not sorted(trap_lifetimes) == trap_lifetimes:
                raise exc.PriorException

            trap_lifetimes = [serial_species.trap_lifetime for serial_species in cti_params.serial_species]

            if not sorted(trap_lifetimes) == trap_lifetimes:
                raise exc.PriorException

        def describe(self, instance):
            return (
                "\nRunning CTI analysis for... \n\n"
                "Parallel CTI: \n"
                "Parallel Species:\n{}\n\n"
                "Parallel CCD\n{}\n\n"
                "Serial CTI: \n"
                "Serial Species:\n{}\n\n"
                "Serial CCD\n{}\n\n".format(instance.parallel_species, instance.parallel_ccd,
                                            instance.serial_species, instance.serial_ccd))


class HyperPhase(Phase):
    """
    Mixin for hyper phases. Extracts noise scaling maps and creates MaskedCIHyperData objects for analysis.
    """
    hyper_noise_scalars = phase_property.PhasePropertyCollection("hyper_noise_scalars")

    def __init__(self, phase_name, phase_tag, phase_folders, *args, **kwargs):
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders, *args, **kwargs)

    def extract_ci_hyper_data(self, data, mask, noise_scaling_maps):
        raise NotImplementedError()

    def noise_scaling_maps_from_result(self, result):
        raise NotImplementedError()

    def make_analysis(self, ci_datas, cti_settings, results=None, pool=None):
        """
        Create an analysis object. Also calls the prior passing and image modifying functions to allow child classes to
        change the behaviour of the phase.

        Parameters
        ----------
        cti_settings
        ci_datas
        pool
        results: [Result]
            The result from the previous phase

        Returns
        -------
        analysis: Analysis
            An analysis object that the non-linear optimizer calls to determine the fit of a set of values
        """
        masks = list(map(lambda data: self.mask_function(shape=data.image.shape), ci_datas))
        noise_scaling_maps = self.noise_scaling_maps_from_result(results[-1])
        ci_datas_fit = [self.extract_ci_hyper_data(data=data, mask=mask, noise_scaling_maps=maps) for
                        data, mask, maps in zip(ci_datas, masks, noise_scaling_maps)]
        ci_datas_full = list(map(lambda data, mask, maps:
                                 ci_data.MaskedCIHyperData(image=data.image, noise_map=data.noise_map,
                                                           ci_pre_cti=data.ci_pre_cti, mask=mask,
                                                           ci_pattern=data.ci_pattern, ci_frame=data.ci_frame,
                                                           noise_scaling_maps=maps),
                                 ci_datas, masks, noise_scaling_maps))

        self.pass_priors(results)

        class HyperAnalysis(self.__class__.Analysis):

            def fit(self, instance):
                """
                Runs the analysis. Determine how well the supplied cti_params fits the image.

                Params
                ----------
                parallel : arctic_params.ParallelParams
                    A class describing the parallel cti model and parameters.
                parallel : arctic_params.SerialParams
                    A class describing the serial cti model and parameters.
                hyp : ci_hyper.HyperCINoise
                    A class describing the noises scaling ci_hyper-parameters.

                Returns
                -------
                result: Result
                    An object comprising the final cti_params instances generated and a corresponding figure_of_merit
                """
                cti_params = cti_params_for_instance(instance=instance)
                self.check_trap_lifetimes_are_ascending(cti_params=cti_params)
                pipe_cti_pass = partial(pipe_cti_hyper, cti_params=cti_params, cti_settings=self.cti_settings,
                                        hyper_noise_scalars=instance.hyper_noise_scalars)
                likelihood = np.sum(list(self.pool.map(pipe_cti_pass, self.ci_datas_extracted)))
                return likelihood

            def describe(self, instance):
                return (
                    "{}"
                    "Hyper Parameters:\n{}\n\n".format(super().describe(instance),
                                                       " ".join(map(str, instance.hyper_noise_scalars))))

        analysis = HyperAnalysis(ci_datas_extracted=ci_datas_fit, ci_datas_full=ci_datas_full,
                                 cti_settings=cti_settings, phase_name=self.phase_name,
                                 results=results, pool=pool)
        return analysis

    @property
    def number_of_noise_scalars(self):
        return len(self.hyper_noise_scalars)

    @number_of_noise_scalars.setter
    def number_of_noise_scalars(self, number):
        self.hyper_noise_scalars = [pm.PriorModel(ci_hyper.CIHyperNoiseScalar) for _ in range(number)]


class ParallelHyperPhase(ParallelPhase, HyperPhase):
    def __init__(self, phase_name, phase_tag=None, phase_folders=None, parallel_species=(), parallel_ccd=None,
                 optimizer_class=nl.MultiNest, mask_function=msk.Mask.empty_for_shape, columns=None,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase with a simple source/CTI model

        Parameters
        ----------
        optimizer_class: class
            The class of a non-linear optimizer
        """
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         parallel_species=parallel_species, parallel_ccd=parallel_ccd,
                         optimizer_class=optimizer_class, mask_function=mask_function, columns=columns,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)
        self.number_of_noise_scalars = 2

    def noise_scaling_maps_from_result(self, result):
        """
        Extract relevant noise scalings maps from the previous result.
        """
        return [result.noise_scaling_maps_of_ci_regions, result.noise_scaling_maps_of_parallel_trails]

    def extract_ci_hyper_data(self, data, mask, noise_scaling_maps):
        return data.parallel_hyper_calibration_data(
            columns=(0, self.columns or data.ci_frame.frame_geometry.parallel_overscan.total_columns),
            mask=mask, noise_scaling_maps=noise_scaling_maps)


class SerialHyperPhase(SerialPhase, HyperPhase):

    def __init__(self, phase_name, phase_tag=None, phase_folders=None, serial_species=(), serial_ccd=None,
                 optimizer_class=nl.MultiNest, mask_function=msk.Mask.empty_for_shape, rows=None,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase with a simple source/CTI model
        """
        HyperPhase.__init__(self=self, phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders)
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         serial_species=serial_species, serial_ccd=serial_ccd, optimizer_class=optimizer_class,
                         mask_function=mask_function, rows=rows,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)
        self.number_of_noise_scalars = 2

    def noise_scaling_maps_from_result(self, result):
        """
        Extract relevant noise scalings maps from the previous result.
        """
        return [result.noise_scaling_maps_of_ci_regions, result.noise_scaling_maps_of_serial_trails]

    def extract_ci_hyper_data(self, data, mask, noise_scaling_maps):
        return data.serial_hyper_calibration_data(rows=self.rows or (0, data.ci_pattern.regions[0].total_rows),
                                                  mask=mask, noise_scaling_maps=noise_scaling_maps)


class ParallelSerialHyperPhase(ParallelSerialPhase, HyperPhase):
    def __init__(self, phase_name, phase_tag=None, phase_folders=None, parallel_species=(), serial_species=(), parallel_ccd=None,
                 serial_ccd=None, optimizer_class=nl.MultiNest, mask_function=msk.Mask.empty_for_shape,
                 cosmic_ray_parallel_buffer=10, cosmic_ray_serial_buffer=10, cosmic_ray_diagonal_buffer=3):
        """
        A phase with a simple source/CTI model

        Parameters
        ----------
        optimizer_class: class
            The class of a non-linear optimizer
        """
        super().__init__(phase_name=phase_name, phase_tag=phase_tag, phase_folders=phase_folders,
                         parallel_species=parallel_species, serial_species=serial_species, parallel_ccd=parallel_ccd,
                         serial_ccd=serial_ccd, optimizer_class=optimizer_class, mask_function=mask_function,
                         cosmic_ray_parallel_buffer=cosmic_ray_parallel_buffer,
                         cosmic_ray_serial_buffer=cosmic_ray_serial_buffer,
                         cosmic_ray_diagonal_buffer=cosmic_ray_diagonal_buffer)
        self.number_of_noise_scalars = 4

    def noise_scaling_maps_from_result(self, result):
        """
        Extract relevant noise scalings maps from the previous result.
        """
        return [result.noise_scaling_maps_of_ci_regions,
                result.noise_scaling_maps_of_parallel_trails,
                result.noise_scaling_maps_of_serial_trails,
                result.noise_scaling_maps_of_serial_overscan_above_trails]

    def extract_ci_hyper_data(self, data, mask, noise_scaling_maps):
        return data.parallel_serial_hyper_calibration_data(mask, noise_scaling_maps=noise_scaling_maps)


def pipe_cti(ci_data_fit, cti_params, cti_settings):
    fit = ci_fit.CIFit(masked_ci_data=ci_data_fit, cti_params=cti_params, cti_settings=cti_settings)
    return fit.figure_of_merit


def pipe_cti_hyper(ci_data_fit, cti_params, cti_settings, hyper_noise_scalars):
    fit = ci_fit.CIHyperFit(masked_hyper_ci_data=ci_data_fit, cti_params=cti_params, cti_settings=cti_settings,
                            hyper_noise_scalars=hyper_noise_scalars)
    return fit.figure_of_merit



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/CIPattern_test.py

Created on: 02/14/18
Author: James Nightingale
"""

import numpy as np
import pytest

from autocti import exc
from autocti.charge_injection import ci_pattern

class TestCIPatternViaList(object):

    def test__2_uniform_patterns__sets_up_collection(self):
        pattern_collection = ci_pattern.uniform_from_lists(normalizations=[1.0, 3.0], regions=[(1, 2, 3, 4)])

        assert type(pattern_collection[0]) == ci_pattern.CIPatternUniform
        assert pattern_collection[0].normalization == 1.0
        assert pattern_collection[0].regions == [(1, 2, 3, 4)]

        assert type(pattern_collection[1]) == ci_pattern.CIPatternUniform
        assert pattern_collection[1].normalization == 3.0
        assert pattern_collection[1].regions == [(1, 2, 3, 4)]

    def test__2_non_uniform_patterns__sets_up_collection(self):
        pattern_collection = ci_pattern.non_uniform_from_lists(normalizations=[1.0, 3.0], regions=[(1, 2, 3, 4)],
                                                               row_slopes=[1.0, 2.0])

        assert type(pattern_collection[0]) == ci_pattern.CIPatternNonUniform
        assert pattern_collection[0].normalization == 1.0
        assert pattern_collection[0].regions == [(1, 2, 3, 4)]
        assert pattern_collection[0].row_slope == 1.0

        assert type(pattern_collection[0]) == ci_pattern.CIPatternNonUniform
        assert pattern_collection[1].normalization == 3.0
        assert pattern_collection[1].regions == [(1, 2, 3, 4)]
        assert pattern_collection[1].row_slope == 2.0


class TestCIPattern(object):
    class TestConstructor:

        def test__setup_all_attributes_correctly(self):
            pattern = ci_pattern.CIPattern(normalization=1.0, regions=[(1, 2, 3, 4)])

            assert pattern.normalization == 1.0
            assert pattern.regions == [(1, 2, 3, 4)]

    class TestCheckImageDimensions:

        def test__pattern_has_more_rows_than_image__1_region(self):
            pattern = ci_pattern.CIPattern(normalization=1.0, regions=([(0, 3, 0, 1)]))

            with pytest.raises(exc.CIPatternException):
                pattern.check_pattern_is_within_image_dimensions(dimensions=(2, 6))

        def test__pattern_has_more_columns_than_image__1_region(self):
            pattern = ci_pattern.CIPattern(normalization=1.0, regions=([(0, 1, 0, 3)]))

            with pytest.raises(exc.CIPatternException):
                pattern.check_pattern_is_within_image_dimensions(dimensions=(6, 2))

        def test__check_rows_and_columns__2_region(self):
            pattern = ci_pattern.CIPattern(normalization=1.0, regions=([(0, 3, 0, 1), (0, 1, 0, 3)]))

            with pytest.raises(exc.CIPatternException):
                pattern.check_pattern_is_within_image_dimensions(dimensions=(2, 6))

            pattern = ci_pattern.CIPattern(normalization=1.0, regions=([(0, 3, 0, 1), (0, 1, 0, 3)]))

            with pytest.raises(exc.CIPatternException):
                pattern.check_pattern_is_within_image_dimensions(dimensions=(6, 2))

        def test__region_cannot_be_negative(self):
            with pytest.raises(exc.RegionException):
                ci_pattern.CIPattern(normalization=1.0, regions=([(-1, 0, 0, 0)]))

            with pytest.raises(exc.RegionException):
                ci_pattern.CIPattern(normalization=1.0, regions=([(0, -1, 0, 0)]))

            with pytest.raises(exc.RegionException):
                ci_pattern.CIPattern(normalization=1.0, regions=([(0, 0, -1, 0)]))

            with pytest.raises(exc.RegionException):
                ci_pattern.CIPattern(normalization=1.0, regions=([(0, 0, 0, -1)]))


class TestCIPatternUniform(object):
    class TestConstructor:

        def test__setup_all_attributes_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=3.0, regions=[(1, 4, 2, 4)])

            assert pattern.normalization == 3.0
            assert pattern.regions == [(1, 4, 2, 4)]

    class TestComputeCIPreCTI:

        def test__uniform_pattern_1_region_normalization_10__correct_ci_pre_cti(self):
            pattern = ci_pattern.CIPatternUniform(normalization=10.0, regions=[(0, 2, 0, 2)])

            ci_pre_cti = pattern.ci_pre_cti_from_shape(shape=(3, 3))

            assert (ci_pre_cti == np.array([[10.0, 10.0, 0.0],
                                            [10.0, 10.0, 0.0],
                                            [0.0, 0.0, 0.0]])).all()

        def test__pattern_bigger_than_image_dimensions__raises_error(self):
            pattern = ci_pattern.CIPatternUniform(normalization=10.0, regions=[(0, 2, 0, 1)])

            with pytest.raises(exc.CIPatternException):
                pattern.ci_pre_cti_from_shape(shape=(1, 1))


class TestCIPatternNonUniform(object):
    class TestConstructor:

        def test__setup_all_attributes_correctly(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=5.0, regions=[(2, 3, 3, 4)], row_slope=0.1)

            assert pattern.normalization == 5.0
            assert pattern.regions == [(2, 3, 3, 4)]
            assert pattern.row_slope == 0.1

    class TestMeanChargeInColumn:

        def test__column_values_all_10__no_mask__measures_mean_as_10(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 1, 0, 1)], row_slope=0.0)

            column = np.array([10.0, 10.0, 10.0])
            column_mask = np.array([False, False, False])

            assert pattern.mean_charge_in_column(column, column_mask) == 10.0

        def test__column_values_non_uniform__mask_omits_2_values__measures_mean_correctly(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 1, 0, 1)], row_slope=0.0)

            column = np.array([11.0, 8.0, 15.0, 10])
            column_mask = np.array([False, True, True, False])

            assert pattern.mean_charge_in_column(column, column_mask) == 10.5

        def test__column_is_fully_masked__mean_is_none(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 1, 0, 1)], row_slope=0.0)

            column = np.array([11.0, 8.0, 15.0, 10])
            column_mask = np.array([True, True, True, True])

            assert pattern.mean_charge_in_column(column, column_mask) is None

        def test__column_values_all_different__no_row_non_uniformity__measures_mean_correctly(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 1, 0, 1)], row_slope=0.0)

            column = np.array([11.0, 8.0, 15.0, 10])
            column_mask = np.array([False, False, False, False])

            assert pattern.mean_charge_in_column(column, column_mask) == 11.0

        def test__column_values_identical_after_row_non_uniformity__measures_mean_as_input_normalization(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)], row_slope=-0.01)

            column = np.array([100.0, 99.3, 98.9])
            column_mask = np.array([False, False, False])

            assert pattern.mean_charge_in_column(column, column_mask) == pytest.approx(100., 1e-4)

        def test__column_values_10_above_normliazation_of_100_after_row_non_uniformity__measures_mean_as_110(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)], row_slope=-0.01)

            column = np.array([110.0, 109.3, 108.9])
            column_mask = np.array([False, False, False])

            assert pattern.mean_charge_in_column(column, column_mask) == pytest.approx(110., 1e-4)

        def test__same_as_above_but_normalization_is_500(self):
            pattern_sim = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=1.0)
            column = pattern_sim.generate_column(size=10, normalization=500.0)

            pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)], row_slope=1.0)

            column_mask = np.array([False] * 10)

            assert pattern.mean_charge_in_column(column, column_mask) == 500.0

        def test__subtract_30_from_the_column_with_normalization_500__mean_is_470_due_to_subtraction(self):
            pattern_sim = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=0.0)

            column = pattern_sim.generate_column(size=10, normalization=500.0)

            pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)], row_slope=0.0)

            column -= 30.0
            column_mask = np.array([False] * 10)

            assert pattern.mean_charge_in_column(column, column_mask) == 470.0

    class TestMeanChargeInAllColumns:

        def test__uniform_column__ci_region_is_entire_image(self):
            ci_column = np.array([[10.0],
                                  [10.0],
                                  [10.0]])

            column_mask = np.array([[False],
                                    [False],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 3, 0, 1)], row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [10.0]

        def test__non_uniform_column__ci_region_is_entire_image(self):
            ci_column = np.array([[13.0],
                                  [10.0],
                                  [10.0]])

            column_mask = np.array([[False],
                                    [False],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 3, 0, 1)], row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [11.0]

        def test__non_uniform_column__mask_changes_mean_estimate(self):
            ci_column = np.array([[13.0],
                                  [10.0],
                                  [10.0]])

            column_mask = np.array([[False],
                                    [True],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 3, 0, 1)], row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [11.5]

        def test__non_uniform_column__fully_masked_mean_estimate_is_none(self):
            ci_column = np.array([[13.0],
                                  [10.0],
                                  [10.0]])

            column_mask = np.array([[True],
                                    [True],
                                    [True]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 3, 0, 1)], row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [None]

        def test__non_uniform_column__ci_region_is_only_two_entries(self):
            ci_column = np.array([[12.0],
                                  [10.0],
                                  [10.0]])

            column_mask = np.array([[False],
                                    [False],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 2, 0, 1)], row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [11.0]

        def test__uniform_column__three_ci_regions(self):
            ci_column = np.array([[10.0],
                                  [10.0],
                                  [10.0],
                                  [0.0],
                                  [10.0],
                                  [10.0],
                                  [10.0],
                                  [0.0],
                                  [10.0],
                                  [10.0],
                                  [10.0]])

            column_mask = np.array([[False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 1), (4, 7, 0, 1), (8, 11, 0, 1)],
                                                     row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [10.0, 10.0, 10.0]

        def test__non_uniform_column__three_ci_regions(self):
            ci_column = np.array([[13.0],
                                  [10.0],
                                  [10.0],
                                  [0.0],
                                  [10.0],
                                  [7.0],
                                  [10.0],
                                  [0.0],
                                  [16.0],
                                  [16.0],
                                  [7.0]])

            column_mask = np.array([[False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 1), (4, 7, 0, 1), (8, 11, 0, 1)],
                                                     row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [11.0, 9.0, 13.0]

        def test__non_uniform_column__three_ci_regions__mask_changes_each_mean(self):
            ci_column = np.array([[13.0],
                                  [10.0],
                                  [10.0],
                                  [0.0],
                                  [10.0],
                                  [7.0],
                                  [10.0],
                                  [0.0],
                                  [16.0],
                                  [16.0],
                                  [7.0]])

            column_mask = np.array([[False],
                                    [True],
                                    [False],
                                    [False],
                                    [True],
                                    [False],
                                    [True],
                                    [False],
                                    [False],
                                    [True],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 1), (4, 7, 0, 1), (8, 11, 0, 1)],
                                                     row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [11.5, 7.0, 11.5]

        def test__non_uniform_column__three_ci_regions__fully_masked_includes_a_none(self):
            ci_column = np.array([[13.0],
                                  [10.0],
                                  [10.0],
                                  [0.0],
                                  [10.0],
                                  [7.0],
                                  [10.0],
                                  [0.0],
                                  [16.0],
                                  [16.0],
                                  [7.0]])

            column_mask = np.array([[False],
                                    [True],
                                    [False],
                                    [False],
                                    [True],
                                    [True],
                                    [True],
                                    [False],
                                    [False],
                                    [True],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 1), (4, 7, 0, 1), (8, 11, 0, 1)],
                                                     row_slope=0.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [11.5, None, 11.5]

        def test__non_uniform_column__three_ci_regions__include_row_non_uniformity(self):
            pattern_sim = ci_pattern.CIPatternNonUniform(normalization=500.0,
                                                         regions=[(0, 1, 0, 1)],
                                                         row_slope=1.0)

            column = pattern_sim.generate_column(size=3, normalization=500.0)

            ci_column = np.zeros((11,))

            ci_column[0:3] = column
            ci_column[4:7] = column
            ci_column[8:11] = column

            column_mask = np.array([[False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False],
                                    [False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=500.0,
                                                     regions=[(0, 3, 0, 1), (4, 7, 0, 1), (8, 11, 0, 1)],
                                                     row_slope=1.0)

            mean_of_columns = pattern.mean_charge_in_all_image_columns(column=ci_column, column_mask=column_mask)

            assert mean_of_columns == [500.0, 500.0, 500.0]

    class TestSetupMockNonUniformImage:

        def test__one_ci_region__all_columns_same__set_up_identical_image(self):
            ci_image = np.array([[10.0, 10.0, 10.0],
                                 [10.0, 10.0, 10.0],
                                 [10.0, 10.0, 10.0]])

            mask = np.array([[False, False, False],
                             [False, False, False],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3)],
                                                     row_slope=0.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == ci_image).all()

        def test__one_ci_region__different_columns_but_same_values__set_up_identical_image(self):
            ci_image = np.array([[10.0, 11.0, 18.0],
                                 [10.0, 11.0, 18.0],
                                 [10.0, 11.0, 18.0]])

            mask = np.array([[False, False, False],
                             [False, False, False],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3)],
                                                     row_slope=0.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == ci_image).all()

        def test__one_ci_region__different_columns_different_values__set_up_as_means(self):
            ci_image = np.array([[10.0, 11.0, 18.0],
                                 [13.0, 8.0, 18.0],
                                 [16.0, 5.0, 18.0]])

            mask = np.array([[False, False, False],
                             [False, False, False],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3)],
                                                     row_slope=0.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == np.array([[13.0, 8.0, 18.0],
                                            [13.0, 8.0, 18.0],
                                            [13.0, 8.0, 18.0]])).all()

        def test__one_ci_region__mask_changes_mean_values(self):
            ci_image = np.array([[10.0, 11.0, 18.0],
                                 [13.0, 8.0, 18.0],
                                 [16.0, 5.0, 18.0]])

            mask = np.array([[True, True, True],
                             [False, False, True],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3)],
                                                     row_slope=0.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == np.array([[14.5, 6.5, 18.0],
                                            [14.5, 6.5, 18.0],
                                            [14.5, 6.5, 18.0]])).all()

        def test__one_ci_region__fully_masked_raises_an_error(self):
            ci_image = np.array([[10.0, 11.0, 18.0],
                                 [13.0, 8.0, 18.0],
                                 [16.0, 5.0, 18.0]])

            mask = np.array([[True, True, True],
                             [False, True, True],
                             [False, True, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3)],
                                                     row_slope=0.0)

            with pytest.raises(exc.CIPatternException):
                pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

        def test__one_ci_region__includes_row_non_uniformity__set_up_as_means(self):
            pattern_sim = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=1.0)

            column = pattern_sim.generate_column(size=3, normalization=500.0)

            ci_image = np.zeros((3, 3))

            ci_image[:, 0] = column
            ci_image[:, 1] = column
            ci_image[:, 2] = column

            mask = np.array([[False, False, False],
                             [False, False, False],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 3, 0, 3)],
                                                     row_slope=1.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == np.array([[500.0, 500.0, 500.0],
                                            [500.0, 500.0, 500.0],
                                            [500.0, 500.0, 500.0]])).all()

        def test__two_ci_regions__takes_mean_of_entire_image_column(self):
            ci_image = np.array([[5.0, 9.0, 8.0],
                                 [5.0, 6.0, 8.0],
                                 [5.0, 3.0, 8.0],
                                 [0.0, 0.0, 0.0],
                                 [5.0, 3.0, 6.0],
                                 [5.0, 6.0, 6.0],
                                 [5.0, 9.0, 6.0]])

            mask = np.array([[False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3), (4, 7, 0, 3)], row_slope=0.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == np.array([[5.0, 6.0, 7.0],
                                            [5.0, 6.0, 7.0],
                                            [5.0, 6.0, 7.0],
                                            [0.0, 0.0, 0.0],
                                            [5.0, 6.0, 7.0],
                                            [5.0, 6.0, 7.0],
                                            [5.0, 6.0, 7.0]])).all()

        def test__two_ci_regions__full_mask_on_one_column__mean_uses_other_column(self):
            ci_image = np.array([[6.0, 9.0, 8.0],
                                 [6.0, 9.0, 8.0],
                                 [6.0, 9.0, 8.0],
                                 [0.0, 0.0, 0.0],
                                 [5.0, 8.0, 6.0],
                                 [5.0, 8.0, 6.0],
                                 [5.0, 8.0, 6.0]])

            mask = np.array([[False, True, False],
                             [False, True, False],
                             [False, True, False],
                             [False, False, False],
                             [True, False, True],
                             [True, False, True],
                             [True, False, True]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0,
                                                     regions=[(0, 3, 0, 3), (4, 7, 0, 3)], row_slope=0.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == np.array([[6.0, 8.0, 8.0],
                                            [6.0, 8.0, 8.0],
                                            [6.0, 8.0, 8.0],
                                            [0.0, 0.0, 0.0],
                                            [6.0, 8.0, 8.0],
                                            [6.0, 8.0, 8.0],
                                            [6.0, 8.0, 8.0]])).all()

        def test__two_ci_regions__includes_row_non_uniformity__set_up_as_means(self):
            pattern_sim = ci_pattern.CIPatternNonUniform(normalization=500.0,
                                                         regions=[(0, 1, 0, 1)],
                                                         row_slope=1.0)

            column = pattern_sim.generate_column(size=3, normalization=500.0)

            ci_image = np.zeros((7, 3))

            ci_image[0:3, 0] = column
            ci_image[0:3, 1] = column
            ci_image[0:3, 2] = column

            ci_image[4:7, 0] = column
            ci_image[4:7, 1] = column
            ci_image[4:7, 2] = column

            mask = np.array([[False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False],
                             [False, False, False]])

            pattern = ci_pattern.CIPatternNonUniform(normalization=500.0,
                                                     regions=[(0, 3, 0, 3), (4, 7, 0, 3)], row_slope=1.0)

            mock_image = pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=ci_image, mask=mask)

            assert (mock_image == np.array([[500.0, 500.0, 500.0],
                                            [500.0, 500.0, 500.0],
                                            [500.0, 500.0, 500.0],
                                            [0.0, 0.0, 0.0],
                                            [500.0, 500.0, 500.0],
                                            [500.0, 500.0, 500.0],
                                            [500.0, 500.0, 500.0]])).all()

        def test__pattern_bigger_than_image_dimensions__raises_error(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 2, 0, 2)], row_slope=0.0)

            with pytest.raises(exc.CIPatternException):
                pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=np.ones((1, 1)), mask=np.ma.ones((1, 1)))


class TestCIPatternSimulateUniform(object):
    class TestSimulateCIPreCTI:

        def test__image_3x3__1_ci_region(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=10.0, regions=[(0, 2, 0, 2)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(3, 3))

            assert (image1 == np.array([[10.0, 10.0, 0.0],
                                        [10.0, 10.0, 0.0],
                                        [0.0, 0.0, 0.0]])).all()

        def test__image_3x3__2_ci_regions(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=10.0,
                                                          regions=[(0, 2, 0, 2), (2, 3, 2, 3)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(3, 3))

            assert (image1 == np.array([[10.0, 10.0, 0.0],
                                        [10.0, 10.0, 0.0],
                                        [0.0, 0.0, 10.0]])).all()

        def test__same_as_above__different_normalizaition(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=20.0,
                                                          regions=[(0, 2, 0, 2), (2, 3, 2, 3)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(3, 3))

            assert (image1 == np.array([[20.0, 20.0, 0.0],
                                        [20.0, 20.0, 0.0],
                                        [0.0, 0.0, 20.0]])).all()

        def test__same_as_above__different_everything(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=30.0,
                                                          regions=[(0, 3, 0, 2), (2, 3, 2, 3)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(4, 3))

            assert (image1 == np.array([[30.0, 30.0, 0.0],
                                        [30.0, 30.0, 0.0],
                                        [30.0, 30.0, 30.0],
                                        [0.0, 0.0, 0.0]])).all()

        def test__pattern_bigger_than_image_dimensions__raises_error(self):
            pattern = ci_pattern.CIPatternUniform(normalization=10.0, regions=[(0, 2, 0, 2)])

            with pytest.raises(exc.CIPatternException):
                pattern.ci_pre_cti_from_shape(shape=(1, 1))


class TestCIPatternSimulateNonUniform(object):
    class TestSimulateRegion:

        def test__uniform_column_and_uniform_row__returns_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=0.0)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(3, 3), ci_seed=1)

            assert (non_uniform_charge_region == np.array([[100.0, 100.0, 100.0],
                                                           [100.0, 100.0, 100.0],
                                                           [100.0, 100.0, 100.0]])).all()

        def test__same_as_above_but_different_normalization_and_dimensions__returns_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=0.0)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(5, 3), ci_seed=1)

            assert (non_uniform_charge_region == np.array([[500.0, 500.0, 500.0],
                                                           [500.0, 500.0, 500.0],
                                                           [500.0, 500.0, 500.0],
                                                           [500.0, 500.0, 500.0],
                                                           [500.0, 500.0, 500.0]])).all()

        def test__non_uniform_column_and_uniform_row__returns_non_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=0.0)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(3, 3), ci_seed=1,
                                                                    column_deviation=1.0)

            non_uniform_charge_region = np.round(non_uniform_charge_region, 1)

            assert (non_uniform_charge_region == np.array([[101.6, 99.4, 99.5],
                                                           [101.6, 99.4, 99.5],
                                                           [101.6, 99.4, 99.5]])).all()

        def test__same_as_above_but_different_normalization_and_dimensions_2__returns_non_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=0.0)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(5, 3), ci_seed=1,
                                                                    column_deviation=1.0)

            non_uniform_charge_region = np.round(non_uniform_charge_region, 1)

            assert (non_uniform_charge_region == np.array([[501.6, 499.4, 499.5],
                                                           [501.6, 499.4, 499.5],
                                                           [501.6, 499.4, 499.5],
                                                           [501.6, 499.4, 499.5],
                                                           [501.6, 499.4, 499.5]])).all()

        def test__uniform_column_and_non_uniform_row__returns_non_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=-0.01)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(3, 3), ci_seed=1)

            non_uniform_charge_region = np.round(non_uniform_charge_region, 1)

            assert (non_uniform_charge_region == np.array([[100.0, 100.0, 100.0],
                                                           [99.3, 99.3, 99.3],
                                                           [98.9, 98.9, 98.9]])).all()

        def test__same_as_above_but_different_normalization_and_dimensions_3__returns_non_uniform_charge(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=-0.01)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(5, 3), ci_seed=1)

            non_uniform_charge_region = np.round(non_uniform_charge_region, 1)

            assert (non_uniform_charge_region == np.array([[500.0, 500.0, 500.0],
                                                           [496.5, 496.5, 496.5],
                                                           [494.5, 494.5, 494.5],
                                                           [493.1, 493.1, 493.1],
                                                           [492.0, 492.0, 492.0]])).all()

        def test__non_uniform_column_and_non_uniform_row__returns_non_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=-0.01)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(3, 3), ci_seed=1,
                                                                    column_deviation=1.0)

            non_uniform_charge_region = np.round(non_uniform_charge_region, 1)

            assert (non_uniform_charge_region == np.array([[101.6, 99.4, 99.5],
                                                           [100.9, 98.7, 98.8],
                                                           [100.5, 98.3, 98.4]])).all()

        def test__same_as_above_but_different_normalization_and_dimensions_4__returns_non_uniform_charge_region(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=500.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=-0.01)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(5, 3), ci_seed=1,
                                                                    column_deviation=1.0)

            non_uniform_charge_region = np.round(non_uniform_charge_region, 1)

            assert (non_uniform_charge_region == np.array([[501.6, 499.4, 499.5],
                                                           [498.2, 495.9, 496.0],
                                                           [496.1, 493.9, 494.0],
                                                           [494.7, 492.5, 492.6],
                                                           [493.6, 491.4, 491.5]])).all()

        def test__non_uniform_columns_with_large_deviation_value__no_negative_charge_columns_are_generated(self):
            sim_pattern = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 1, 0, 1)],
                                                         row_slope=0.0)

            non_uniform_charge_region = sim_pattern.simulate_region(region_dimensions=(10, 10), ci_seed=1,
                                                                    column_deviation=100.0)

            assert (non_uniform_charge_region > 0).all()

    class TestSimulateImage:

        def test__no_non_uniformity__identical_to_uniform_image__one_ci_region(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=10.0, regions=[(2, 4, 0, 5)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(5, 5))

            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(2, 4, 0, 5)],
                                                                 row_slope=0.0)
            image2 = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5))

            assert (image1 == image2).all()

        def test__no_non_uniformity__identical_to_uniform_image__rectangular_image(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=10.0, regions=[(2, 4, 0, 5)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(5, 7))

            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(
                normalization=10.0,
                regions=[(2, 4, 0, 5)],
                row_slope=0.0)
            image2 = sim_pattern_non_uni.simulate_ci_pre_cti(column_deviation=0.0,
                                                             shape=(5, 7))

            assert (image1 == image2).all()

        def test__no_non_uniformity__identical_to_uniform_image__different_normalization_and_region(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=100.0, regions=[(1, 4, 2, 5)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(5, 5))

            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(1, 4, 2, 5)],
                                                                 row_slope=0.0)
            image2 = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5))

            assert (image1 == image2).all()

        def test__no_non_uniformity__identical_to_uniform_image__two_ci_regions(self):
            sim_pattern_uni = ci_pattern.CIPatternUniform(normalization=100.0,
                                                          regions=[(0, 2, 0, 2), (2, 3, 0, 5)])
            image1 = sim_pattern_uni.simulate_ci_pre_cti(shape=(5, 5))

            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0,
                                                                 regions=[(0, 2, 0, 2), (2, 3, 0, 5)],
                                                                 row_slope=0.0)
            image2 = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5))

            assert (image1 == image2).all()

        def test__non_uniformity_in_columns_only__one_ci_region__image_is_correct(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 3, 0, 3)],
                                                                 row_slope=0.0)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5), ci_seed=1,
                                                            column_deviation=1.0)

            image = np.round(image, 1)

            assert (image == np.array([[101.6, 99.4, 99.5, 0.0, 0.0],
                                       [101.6, 99.4, 99.5, 0.0, 0.0],
                                       [101.6, 99.4, 99.5, 0.0, 0.0],
                                       [0.0, 0.0, 0.0, 0.0, 0.0],
                                       [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        def test__non_uniformity_in_columns_only__different_ci_region_to_above__image_is_correct(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(1, 4, 1, 4)],
                                                                 row_slope=0.0)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5), ci_seed=1,
                                                            column_deviation=1.0)

            image = np.round(image, 1)

            assert (image == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                       [0.0, 101.6, 99.4, 99.5, 0.0],
                                       [0.0, 101.6, 99.4, 99.5, 0.0],
                                       [0.0, 101.6, 99.4, 99.5, 0.0],
                                       [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        def test__non_uniformity_in_columns_only__two_ci_regions__image_is_correct(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0,
                                                                 regions=[(1, 4, 1, 3), (1, 4, 4, 5)],
                                                                 row_slope=0.0)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5), ci_seed=1,
                                                            column_deviation=1.0)

            image = np.round(image, 1)

            assert (image == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                       [0.0, 101.6, 99.4, 0.0, 101.6],
                                       [0.0, 101.6, 99.4, 0.0, 101.6],
                                       [0.0, 101.6, 99.4, 0.0, 101.6],
                                       [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        def test__non_uniformity_in_columns_only__maximum_normalization_input__does_not_simulate_above(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0, regions=[(0, 5, 0, 5)],
                                                                 row_slope=0.0)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5), ci_seed=1,
                                                            column_deviation=100.0, maximum_normalization=100.0)

            image = np.round(image, 1)

            # Checked ci_seed to ensure the max is above 100.0 without a maximum_normalization
            assert np.max(image) < 100.0

        def test__non_uniformity_in_rows_only__one_ci_region__image_is_correct(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0,
                                                                 regions=[(0, 3, 0, 3)],
                                                                 row_slope=-0.01)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(column_deviation=0.0,
                                                            shape=(5, 5))

            image = np.round(image, 1)

            assert (image == np.array([[100.0, 100.0, 100.0, 0.0, 0.0],
                                       [99.3, 99.3, 99.3, 0.0, 0.0],
                                       [98.9, 98.9, 98.9, 0.0, 0.0],
                                       [0.0, 0.0, 0.0, 0.0, 0.0],
                                       [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

        def test__non_uniformity_in_rows_only__two_ci_regions__image_is_correct(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0,
                                                                 regions=[(1, 5, 1, 4), (0, 5, 4, 5)],
                                                                 row_slope=-0.01)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5))

            image = np.round(image, 1)

            assert (image == np.array([[0.0, 0.0, 0.0, 0.0, 100.0],
                                       [0.0, 100.0, 100.0, 100.0, 99.3],
                                       [0.0, 99.3, 99.3, 99.3, 98.9],
                                       [0.0, 98.9, 98.9, 98.9, 98.6],
                                       [0.0, 98.6, 98.6, 98.6, 98.4]])).all()

        def test__non_uniformity_in_rows_and_columns__two_ci_regions__image_is_correct(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0,
                                                                 regions=[(1, 5, 1, 4), (0, 5, 4, 5)],
                                                                 row_slope=-0.01)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5), ci_seed=1,
                                                            column_deviation=1.0)

            image = np.round(image, 1)

            assert (image == np.array([[0.0, 0.0, 0.0, 0.0, 101.6],
                                       [0.0, 101.6, 99.4, 99.5, 100.9],
                                       [0.0, 100.9, 98.7, 98.8, 100.5],
                                       [0.0, 100.5, 98.3, 98.4, 100.2],
                                       [0.0, 100.2, 98.0, 98.1, 100.0]])).all()

        def test__non_uniformity_in_rows_and_columns__no_random_seed__two_ci_regions_are_identical(self):
            sim_pattern_non_uni = ci_pattern.CIPatternNonUniform(normalization=100.0,
                                                                 regions=[(0, 2, 0, 5), (3, 5, 0, 5)],
                                                                 row_slope=-0.01)

            image = sim_pattern_non_uni.simulate_ci_pre_cti(shape=(5, 5),
                                                            column_deviation=1.0)

            image = np.round(image, 1)

            assert (image[0:2, 0:5] == image[3:5, 0:5]).all()

        def test__pattern_bigger_than_image_dimensions__raises_error(self):
            pattern = ci_pattern.CIPatternNonUniform(normalization=10.0, regions=[(0, 2, 0, 2)],
                                                     row_slope=0.0)

            with pytest.raises(exc.CIPatternException):
                pattern.ci_pre_cti_from_ci_image_and_mask(ci_image=np.ones((1, 1)), mask=np.ma.ones((1, 1)))



from autofit import conf
from matplotlib import pyplot as plt

from autocti.charge_injection.plotters import data_plotters
from autocti.data.plotters import plotter_util


def plot_ci_subplot(
        ci_data, mask=None, extract_array_from_mask=False,
        figsize=None, aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        titlesize=10, xlabelsize=10, ylabelsize=10, xyticksize=10,
        output_path=None, output_filename='ci_data', output_format='show'):
    """Plot the ci data as a sub-plot of all its quantites (e.g. the data, noise_map-map, PSF, Signal-to_noise-map, \
     etc).

    Set *autolens.data.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ci_data : data.CCDData
        The ci data, which includes the observed data, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or data.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the data, this plots those pixels \
        over the immage.
    ignore_config : bool
        If *False*, the config file general.ini is used to determine whether the subpot is plotted. If *True*, the \
        config file is ignored.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(number_subplots=4)

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_image(
        ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format)

    plt.subplot(rows, columns, 2)

    plot_noise_map(
        ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format)

    plt.subplot(rows, columns, 3)

    plot_ci_pre_cti(
        ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format)

    plt.subplot(rows, columns, 4)

    plot_signal_to_noise_map(
        ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=True,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format)

    plotter_util.output_subplot_array(output_path=output_path, output_filename=output_filename,
                                      output_format=output_format)

    plt.close()


def plot_ci_data_individual(ci_data, mask=None, extract_array_from_mask=False,
                            should_plot_image=False,
                            should_plot_noise_map=False,
                            should_plot_ci_pre_cti=False,
                            should_plot_signal_to_noise_map=False,
                            output_path=None, output_format='png'):
    """Plot each attribute of the ci data as individual figures one by one (e.g. the data, noise_map-map, PSF, \
     Signal-to_noise-map, etc).

    Set *autolens.data.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ci_data : data.CCDData
        The ci data, which includes the observed data, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the data's coordinate system is plotted as a 'x'.
    """

    if should_plot_image:
        plot_image(ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask,
                   output_path=output_path, output_format=output_format)

    if should_plot_noise_map:
        plot_noise_map(ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask,
                       output_path=output_path, output_format=output_format)

    if should_plot_ci_pre_cti:
        plot_ci_pre_cti(ci_data=ci_data, extract_array_from_mask=extract_array_from_mask,
                        output_path=output_path, output_format=output_format)

    if should_plot_signal_to_noise_map:
        plot_signal_to_noise_map(ci_data=ci_data, mask=mask, extract_array_from_mask=extract_array_from_mask,
                                 output_path=output_path, output_format=output_format)


def plot_image(
        ci_data, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Charge Injection Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='ci_image'):
    """Plot the observed image of the ci data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : CIFrame
        The image of the data.
    """
    data_plotters.plot_image(
        image=ci_data.image, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
        linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_noise_map(
        ci_data, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Charge Injection Noise-Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='ci_noise_map'):
    """Plot the observed noise_map of the ci data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    noise_map : CIFrame
        The noise map of the data.
    """
    data_plotters.plot_noise_map(
        noise_map=ci_data.noise_map, mask=mask, extract_array_from_mask=extract_array_from_mask, as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
        linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_ci_pre_cti(
        ci_data, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Charge Injection Pre-CTI Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='ci_pre_cti'):
    """Plot the observed ci_pre_cti of the ci data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the data.
    """
    data_plotters.plot_ci_pre_cti(
        ci_pre_cti=ci_data.ci_pre_cti, mask=mask, extract_array_from_mask=extract_array_from_mask,
        as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
        linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)


def plot_signal_to_noise_map(
        ci_data, mask=None, extract_array_from_mask=False, as_subplot=False,
        figsize=(7, 7), aspect='equal',
        cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
        cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
        title='Charge Injection Signal-to-Noise Map', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
        output_path=None, output_format='show', output_filename='ci_signal_to_noise_map'):
    """Plot the observed signal_to_noise_map of the ci data.

    Set *autocti.data.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    signal_to_noise_map : CIFrame
        The signal-to-noise map of the data.
    """
    data_plotters.plot_signal_to_noise_map(
        signal_to_noise_map=ci_data.signal_to_noise_map, mask=mask, extract_array_from_mask=extract_array_from_mask,
        as_subplot=as_subplot,
        figsize=figsize, aspect=aspect,
        cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
        linthresh=linthresh, linscale=linscale,
        cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
        cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels,
        title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize,
        output_path=output_path, output_format=output_format, output_filename=output_filename)



"""
@file python/ELViS_lib/fitslib.py
@date 07/13/16
@author user
"""

# Library : fits files

import astropy.io.fits as pyfits
import numpy
import numpy.ma as MA


def get_extlist_image(filename):
    """Get the list of ImageHDU extensions"""

    ext = pyfits.info(filename, output=False)
    next = len(ext)
    imlist = []
    for iext,ext in enumerate(ext):
        if ext[2] == "ImageHDU":
            imlist.append(iext)
    return next, imlist


def narray2fits(array, outname):
    """ Create a fits file from a narray """

    #try:
    hdu = pyfits.PrimaryHDU(array)
    hdulist = pyfits.HDUList([hdu])
    hdulist.writeto(outname)
    #except:
    #	return 1

    return 0


def MA2fits(MAarray, outname):
    """ Create a fits file from a Masked array """

    arr = numpy.array(MA.filled(MAarray, value=0))
    res = narray2fits(arr, outname)
    return res


def narrayfits_header(array, header, outname):
    """ Create a fits file from a narray with a list of header keywords """

    # Create the HDU
    hdu = pyfits.PrimaryHDU(array)
    hdulist = pyfits.HDUList([hdu])

    # Add the keywords
    for k, v in header.iteritems():
        hdulist[0].header.update(k, v)

    # Write the file
    hdulist.writeto(outname)


def isfits(file0):
    try:
        pyobj = pyfits.open(file0)
    except IOError:
        return False
    pyobj.close()
    return True



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
import numpy as np
from autofit.tools import fit

from autocti.charge_injection import ci_data


class AbstractCIFit(fit.DataFit):

    def __init__(self, masked_ci_data: ci_data.MaskedCIData, noise_map, cti_params, cti_settings):
        """Abstract fit of a charge injection dataset with a model cti image.

        Parameters
        -----------
        masked_ci_data : ci_data.CIDataFit
            The charge injection image that is fitted.
        cti_params : arctic_params.ArcticParams
            The cti model parameters which describe how CTI during clocking.
        cti_settings : arctic_settings.ArcticSettings
            The settings that control how arctic models CTI.
        """
        model_data = masked_ci_data.ci_frame.frame_geometry.add_cti(image=masked_ci_data.ci_pre_cti,
                                                                    cti_params=cti_params,
                                                                    cti_settings=cti_settings)

        super().__init__(data=masked_ci_data.image, noise_map=noise_map,
                         mask=masked_ci_data.mask, model_data=model_data)
        self.ci_data_fit = masked_ci_data
        self.cti_params = cti_params
        self.cti_settings = cti_settings

    @property
    def ci_post_cti(self):
        return self.model_data

    @property
    def ci_pre_cti(self):
        return self.ci_data_fit.ci_pre_cti

    @property
    def image(self):
        return self.data

    @property
    def model_image(self):
        return self.model_data

    @property
    def figure_of_merit(self):
        return self.likelihood


class CIFit(AbstractCIFit):

    def __init__(self, masked_ci_data: ci_data.MaskedCIData, cti_params, cti_settings):
        super().__init__(masked_ci_data, masked_ci_data.noise_map, cti_params, cti_settings)

    @property
    def noise_scaling_map_of_ci_regions(self):
        return self.ci_data_fit.chinj.ci_regions_from_array(array=self.chi_squared_map.copy())

    @property
    def noise_scaling_map_of_parallel_trails(self):
        return self.ci_data_fit.chinj.parallel_non_ci_regions_frame_from_frame(array=self.chi_squared_map.copy())

    @property
    def noise_scaling_map_of_serial_trails(self):
        return self.ci_data_fit.chinj.serial_all_trails_frame_from_frame(array=self.chi_squared_map.copy())

    @property
    def noise_scaling_map_of_serial_overscan_above_trails(self):
        return self.ci_data_fit.chinj.serial_overscan_above_trails_frame_from_frame(array=self.chi_squared_map.copy())


class CIHyperFit(AbstractCIFit):

    def __init__(self, masked_hyper_ci_data: ci_data.MaskedCIHyperData, cti_params, cti_settings, hyper_noise_scalars):
        """Fit a charge injection ci_data-set with a model cti image, also scalng the noises within a Bayesian
        framework.

        Parameters
        -----------
        masked_hyper_ci_data
            The charge injection image that is fitted.
        cti_params : arctic_params.ArcticParams
            The cti model parameters which describe how CTI during clocking.
        cti_settings : arctic_settings.ArcticSettings
            The settings that control how arctic models CTI.
        hyper_noise_scalars :
            The ci_hyper-parameter(s) which the noise_scaling_maps is multiplied by to scale the noise-map.
        """
        self.hyper_noises = hyper_noise_scalars
        self.hyper_noise_map = hyper_noise_map_from_noise_map_and_noise_scalings(
            noise_scaling_maps=masked_hyper_ci_data.noise_scaling_maps, hyper_noise_scalars=hyper_noise_scalars,
            noise_map=masked_hyper_ci_data.noise_map)
        super().__init__(masked_ci_data=masked_hyper_ci_data, noise_map=self.hyper_noise_map, cti_params=cti_params,
                         cti_settings=cti_settings)


def hyper_noise_map_from_noise_map_and_noise_scalings(noise_scaling_maps, hyper_noise_scalars, noise_map):
    """For a noise-map, use the model hyper noise and noise-scaling maps to compute a scaled noise-map.

    Parameters
    -----------
    hyper_noise_scalars
    noise_scaling_maps
    noise_map : imaging.NoiseMap or ndarray
        An array describing the RMS standard deviation error in each pixel, preferably in units of electrons per
        second.
    """
    hyper_noise_maps = list(map(lambda hyper_noise_scalar, noise_scaling_map:
                                hyper_noise_scalar.scaled_noise_map_from_noise_scaling(noise_scaling_map),
                                hyper_noise_scalars, noise_scaling_maps))

    return np.add(noise_map, sum(hyper_noise_maps))



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: python/SHE_ArCTIC/ArcticSettings.py

Created on: 02/13/18
Author: James Nightingale
"""


class ArcticSettings(object):

    def __init__(self, neomode='NEO', parallel=None, serial=None):
        """Sets up the CTI settings for parallel and serial clocking, specified using the \
        ArcticSettings.Settings and ArcticSettings.Settings classes.
        
        Params
        ----------
        neomode : str
            Which function arctic uses for charge clocking. Arctic has multiple clocking algorithms which are \
            numerically equivalent. The 'NEO' algorithm is currently (25/09/17) recommended.
            *Classic* - clocks charge using Massey et al 2014 algorithim without efficiency tricks.
            *NEO* - clocks charge using a trap accounting scheme to speed up code (~50x)
        parallel : ArcticSetup.Settings
            The arctic settings used for clocking in the parallel direction.
        serial : :ArcticSetup.Settings
            The arctic settings used for clocking in the serial direction.
        """
        self.neomode = neomode

        self.parallel = parallel
        self.serial = serial

    def update_fits_header_info(self, ext_header):
        """Output the Arctic Settings into the fits header of a fits image.

        Params
        -----------
        ext_header : astropy.io.hdulist
            The opened header of the astropy fits header.
        """

        if self.parallel is not None:
            ext_header.set('cte_pite', self.parallel.niter, 'Iterations Used In Correction (Parallel)')
            ext_header.set('cte_pwld', self.parallel.well_depth, 'CCD Well Depth (Parallel)')
            ext_header.set('cte_pnts', self.parallel.n_levels, 'Number of levels (Parallel)')

        if self.serial is not None:
            ext_header.set('cte_site', self.serial.niter, 'Iterations Used In Correction (Serial)')
            ext_header.set('cte_swld', self.serial.well_depth, 'CCD Well Depth (Serial)')
            ext_header.set('cte_snts', self.serial.n_levels, 'Number of levels (Serial)')

        return ext_header


class Settings(object):

    def __init__(self, well_depth, niter, express, n_levels, charge_injection_mode=False, readout_offset=0):
        """
        The CTI settings for parallel clocking.

        Returns
        --------
        well_depth : int
            The full well depth of the CCD.
        niter : int
            If CTI is being corrected, niter determines the number of times clocking is run to perform the \
            correction via forward modeling. For adding CTI only one run is required and niter is ignored.
        express : int
            The factor by which pixel-to-pixel transfers are combined for efficiency.
        n_levels : int
            Relic of old arctic code, not used anymore and will be removed in future.
        charge_injection_mode : bool
            If True, clocking is performed in charge injection line mode, where each pixel is clocked and therefore \
             trailed by traps over the entire CCD (as opposed to its distance from the CCD register).
        readout_offset : int
            Introduces an offset which increases the number of transfers each pixel takes in the parallel direction.
        """
        self.well_depth = well_depth
        self.niter = niter
        self.express = express
        self.n_levels = n_levels
        self.charge_injection_mode = charge_injection_mode
        self.readout_offset = readout_offset

    def update_fits_header_info(self, ext_header):
        """Update a fits header to include the parallel CTI settings.

        Params
        -----------
        ext_header : astropy.io.hdulist
            The opened header of the astropy fits header.
        """
        ext_header.set('cte_pite', self.niter, 'Iterations Used In Correction (Parallel)')
        ext_header.set('cte_pwld', self.well_depth, 'CCD Well Depth (Parallel)')
        ext_header.set('cte_pnts', self.n_levels, 'Number of levels (Parallel)')
        return ext_header



import os

from autofit import conf
from autofit.mapper import prior_model
from autofit.optimize import non_linear as nl

from autocti.model import arctic_params
from autocti.model import arctic_settings
from autocti.pipeline import phase as ph
from autocti.pipeline import pipeline as pl
from test.simulation import simulation_util
from test.integration import integration_util

test_type = 'parallel_and_serial'
test_name = 'x3_species_x1_image_no_pool'

test_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + 'output/'
config_path = test_path + 'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    serial_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings, serial=serial_settings)
    data = simulation_util.load_test_ci_data(ci_data_type='ci_uniform', ci_data_model='parallel_and_serial_x3_species',
                                             ci_data_resolution='patch',normalization=84700.0)
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(ci_datas=[data], cti_settings=cti_settings)

def make_pipeline(test_name):
    class ParallelSerialPhase(ph.ParallelSerialPhase):

        def pass_priors(self, results):
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = ParallelSerialPhase(phase_name='phase_1', phase_folders=[test_type, test_name],
                                 optimizer_class=nl.MultiNest,
                                 parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species)],
                                 parallel_ccd=arctic_params.CCD,
                                 serial_species=[prior_model.PriorModel(arctic_params.Species),
                                                 prior_model.PriorModel(arctic_params.Species),
                                                 prior_model.PriorModel(arctic_params.Species)],
                                 serial_ccd=arctic_params.CCD)

    phase1.optimizer.n_live_points = 60
    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.sampling_efficiency = 0.2

    return pl.Pipeline(test_type, phase1)


if __name__ == "__main__":
    pipeline()



import os
import shutil

dirpath = os.path.dirname(os.path.realpath(__file__))

def reset_paths(test_name, output_path):

    try:
        shutil.rmtree(output_path + '/' + test_name)
    except FileNotFoundError:
        pass



class CIHyperNoiseScalar(float):
    def __new__(cls, scale_factor):
        return super().__new__(cls, scale_factor)

    def __init__(self, scale_factor=0.0):
        """The ci_hyper-parameter factor by which the noises is scaled when included in the model-fitting process"""
        float.__init__(scale_factor)

    def scaled_noise_map_from_noise_scaling(self, noise_scaling):
        """Compute the scaled noises map, by multiplying the noises-scaling image by the ci_hyper-parameter factor."""
        return self * noise_scaling

    def __repr__(self):
        return "Noise Scale Factor: {}".format(self) + '\n'



from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior_model
from autocti.data import mask as msk
from autocti.pipeline import pipeline as pl
from autocti.pipeline import phase as ph
from autocti.model import arctic_params

# In this pipeline, we'll perform a basic analysis which fits a single parallel trap species to a set of charge
# injection imaging data. This will include a hyper-phase which scales the noise in the analysis, to prevent
# over-fitting the highest S/N charge injection images. The pipeline uses three phases:

# Phase 1) Fit a small section (the left-most 60 columns) of the charge injection images using a parallel CTI model
#          with 1 trap species and a model for the parallel CCD volume filling parameters.

# Phase 1h) Use the best-fit model from phase 1 to scale the noise of each image, to ensure that the higher and
#          lower S/N images are weighted more equally in their contribution to the likelihood.

# Phase 2) Refit the phase 1 model, using priors initialized from the results of phase 1 and the scaled noise-map
#          computed in phase 2.

def make_pipeline(phase_folders=None, mask_function=msk.Mask.empty_for_shape):

    pipeline_name = 'pipeline_parallel_x1_species'

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the data with a one species parallel CTI model and parallel CCD filling model. In this
    # phase we will:

    # 1) Extract and fit the 60 columns of charge injection imaging data closest to the read-out register (and
    # therefore least affected by serial CTI).

    class ParallelPhase(ph.ParallelPhase):

        def pass_priors(self, results):

            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

    phase1 = ParallelPhase(phase_name='phase_1_init', phase_folders=phase_folders,
                           optimizer_class=nl.MultiNest, mask_function=mask_function,
                           parallel_species=[prior_model.PriorModel(arctic_params.Species)],
                           parallel_ccd=arctic_params.CCD, columns=60)

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster.

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # The best fit model of phase 1 is used to create a 'noise-scaling' map for every charge injection image. These
    # noise-scaling maps are used in conjuction with 'hyper-noise' models to scale the noise-maps in a way that
    # equally weights the fit across all charge injection images.

    # 1) Extract and fit the 60 columns of charge injection imaging data closest to the read-out register (and
    # therefore least affected by serial CTI).

    class ParallelHyperModelFixedPhase(ph.ParallelHyperPhase):

        def pass_priors(self, results):

            self.parallel_species = results.from_phase('phase_1_init').constant.parallel_species
            self.parallel_ccd = results.from_phase('phase_1_init').constant.parallel_ccd

    phase2 = ParallelHyperModelFixedPhase(phase_name='phase_2_noise_scaling', phase_folders=phase_folders,
                                          parallel_species=[prior_model.PriorModel(arctic_params.Species)],
                                          parallel_ccd=arctic_params.CCD,
                                          optimizer_class=nl.MultiNest, columns=None)

    ### PHASE 3 ###

    # In phase 3, we will fit the data with a one species parallel CTI model and parallel CCD filling model. In this
    # phase we will:

    # 1) Use the complete charge injection image, as opposed to extracting a sub-set of columns.
    # 2) Use the scaled noise-map computed in phase 1h.
    # 3) Initialize the priors on the parallel CTI model from the results of phase 1.

    class ParallelHyperFixedPhase(ph.ParallelHyperPhase):

        def pass_priors(self, results):

            self.hyper_noise_scalars = results.from_phase('phase_2_noise_scaling').constant.hyper_noise_scalars
            self.parallel_species = results.from_phase('phase_1_init').variable.parallel_species
            self.parallel_ccd = results.from_phase('phase_1_init').variable.parallel_ccd
            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

    phase3 = ParallelHyperFixedPhase(phase_name='phase_3_final', phase_folders=phase_folders,
                                     optimizer_class=nl.MultiNest)

    # For the final CTI model, constant efficiency mode has a tendancy to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase3.optimizer.const_efficiency_mode = False
    phase3.optimizer.n_live_points = 50
    phase3.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(pipeline_name, phase1, phase2, phase3)


import os

from autofit import conf
from autofit.optimize import non_linear as nl
from autofit.mapper import prior_model
from autocti.model import arctic_params
from autocti.model import arctic_settings

from autocti.pipeline import phase as ph
from autocti.pipeline import pipeline as pl

from test.simulation import simulation_util
from test.integration import integration_util

from multiprocessing import Pool

test_type = 'parallel'
test_name = 'x3_species_x2_images_use_pool'

test_path = '{}/../../'.format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + 'output/'
config_path = test_path + 'config'
conf.instance = conf.Config(config_path=config_path, output_path=output_path)

def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)

    parallel_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=False, readout_offset=0)
    cti_settings = arctic_settings.ArcticSettings(parallel=parallel_settings)
    data = list(map(lambda normalization :
                    simulation_util.load_test_ci_data(ci_data_type='ci_uniform', ci_data_model='parallel_x3_species',
                                                      ci_data_resolution='patch', normalization=normalization),
                    [1000.0, 84700.0]))
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(ci_datas=data, cti_settings=cti_settings, pool=Pool(processes=2))

def make_pipeline(test_name):

    class ParallelPhase(ph.ParallelPhase):

        def pass_priors(self, results):

            self.parallel_ccd.well_fill_alpha = 1.0
            self.parallel_ccd.well_fill_gamma = 0.0

    phase1 = ParallelPhase(phase_name='phase_1', phase_folders=[test_type, test_name],
                           optimizer_class=nl.MultiNest, parallel_species=[prior_model.PriorModel(arctic_params.Species),
                                                                           prior_model.PriorModel(arctic_params.Species),
                                                                           prior_model.PriorModel(arctic_params.Species)],
                           parallel_ccd=arctic_params.CCD,
                           columns=3)

    phase1.optimizer.n_live_points = 60
    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.sampling_efficiency = 0.2

    return pl.Pipeline(test_type, phase1)


if __name__ == "__main__":
    pipeline()



from autofit.tools import path_util
from autofit.optimize import non_linear as nl
from autofit.mapper import prior
from autofit.mapper import prior_model
from autocti.pipeline import pipeline as pl
from autocti.pipeline import phase as ph
from autocti.model import arctic_params

# In this pipeline, we'll perform an analysis which fits three serial trap species to a set of charge
# injection imaging data. This will include a hyper-phase which scales the noise in the analysis, to prevent
# over-fitting the highest S/N charge injection images. The pipeline uses four phases:

# Phase 1) Fit a small section (10 rows of every charge injection) using a serial CTI model
#          with 1 trap species and a model for the serial CCD volume filling parameters.

# Phase 2) Fit a small section (again, 10 rows) using a serial CTI model with 3 trap species and a model for the serial
#          CCD volume filling parameters. The priors on trap densities and volume filling parameters are initialized
#          from the results of phase 1.

# Phase 3) Use the best-fit model from phase 2 to scale the noise of each image, to ensure that the higher and
#          lower S/N images are weighted more equally in their contribution to the likelihood.

# Phase 4) Refit the phase 2 model, using priors initialized from the results of phase 2 and the scaled noise-map
#          computed in phase 3.

def make_pipeline(phase_folders=None):

    pipeline_name = 'pipeline_serial_x3_species_prior_2'
    
    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/'
    phase_folders = path_util.phase_folders_from_phase_folders_and_pipeline_name(phase_folders=phase_folders,
                                                                                pipeline_name=pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the data with a one species serial CTI model and serial CCD filling model. In this
    # phase we will:

    # 1) Extract and fit the 10 rows of each charge injection region which is furthest from the clocking direction (and
    # therefore least affected by parallel CTI).

    class SerialPhase(ph.SerialPhase):

        def pass_priors(self, results):
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase1 = SerialPhase(phase_name='phase_1_x1_species', phase_folders=phase_folders,
                         optimizer_class=nl.MultiNest,
                         serial_species=[prior_model.PriorModel(arctic_params.Species)],
                         serial_ccd=arctic_params.CCD, rows=(0,10))

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster.

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 30
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the data with a three species serial CTI model and serial CCD filling model. In this
    # phase we will:

    # 1) As in phase 1, extract and fit the 10 rows of charge injection imaging data closest to the read-out
    # register.
    # 2) Use priors on the trap density and ccd volume filling parameters based on the results of phase 1.

    class SerialPhase(ph.SerialPhase):

        def pass_priors(self, results):

            previous_total_density = results[-1].constant.serial_species[0].trap_density

            self.serial_species[0].trap_density = prior.UniformPrior(lower_limit=0.005, upper_limit=0.015)
            self.serial_species[1].trap_density = prior.UniformPrior(lower_limit=0.0, upper_limit=previous_total_density)
            self.serial_species[2].trap_density = prior.UniformPrior(lower_limit=0.0, upper_limit=previous_total_density)
            self.serial_species[0].trap_lifetime = prior.UniformPrior(lower_limit=0.0, upper_limit=30.0)
            self.serial_species[1].trap_lifetime = prior.UniformPrior(lower_limit=0.0, upper_limit=30.0)
            self.serial_species[2].trap_lifetime = prior.UniformPrior(lower_limit=0.0, upper_limit=30.0)

            self.serial_ccd.well_notch_depth = results.from_phase('phase_1_x1_species').variable.serial_ccd.well_notch_depth
            self.serial_ccd.well_fill_beta = results.from_phase('phase_1_x1_species').variable.serial_ccd.well_fill_beta
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase2 = SerialPhase(phase_name='phase_2_x3_species_init', phase_folders=phase_folders,
                         serial_species=[prior_model.PriorModel(arctic_params.Species),
                                         prior_model.PriorModel(arctic_params.Species),
                                         prior_model.PriorModel(arctic_params.Species)],
                         serial_ccd=arctic_params.CCD, rows=(0,10), optimizer_class=nl.MultiNest)

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 30
    phase2.optimizer.sampling_efficiency = 0.2

    ### PHASE 3 ###

    # The best fit model of phase 2 is used to create a 'noise-scaling' map for every charge injection image. These
    # noise-scaling maps are used in conjunction with 'hyper-noise' models to scale the noise-maps in a way that
    # equally weights the fit across all charge injection images.

    class SerialHyperModelFixedPhase(ph.SerialHyperPhase):

        def pass_priors(self, results):

            self.serial_species = results.from_phase('phase_2_x3_species_init').constant.serial_species
            self.serial_ccd = results.from_phase('phase_2_x3_species_init').constant.serial_ccd

    phase3 = SerialHyperModelFixedPhase(phase_name='phase_3_noise_scaling', phase_folders=phase_folders,
                                        serial_species=[prior_model.PriorModel(arctic_params.Species),
                                                        prior_model.PriorModel(arctic_params.Species),
                                                        prior_model.PriorModel(arctic_params.Species)],
                                        serial_ccd=arctic_params.CCD, rows=None,
                                        optimizer_class=nl.MultiNest)

    ### PHASE 4 ###

    # In phase 4, we will fit the data with a 3 species serial CTI model and serial CCD filling model. In this
    # phase we will:

    # 1) Use the complete charge injection image, as opposed to extracting a sub-set of rows.
    # 2) Use the scaled noise-map computed in phase 3.
    # 3) Initialize the priors on the serial CTI model from the results of phase 2.

    class SerialHyperFixedPhase(ph.SerialHyperPhase):

        def pass_priors(self, results):

            self.hyper_noise_scalars = results.from_phase('phase_3_noise_scaling').constant.hyper_noise_scalars
            self.serial_species = results.from_phase('phase_2_x3_species_init').variable.serial_species
            self.serial_ccd = results.from_phase('phase_2_x3_species_init').variable.serial_ccd
            self.serial_species[0].trap_density = prior.UniformPrior(lower_limit=0.005, upper_limit=0.015)
            self.serial_ccd.well_fill_alpha = 1.0
            self.serial_ccd.well_fill_gamma = 0.0

    phase4 = SerialHyperFixedPhase(phase_name='phase_4_final', phase_folders=phase_folders,
                                   serial_species=[prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species),
                                                   prior_model.PriorModel(arctic_params.Species)],
                                   serial_ccd=arctic_params.CCD, rows=None,
                                   optimizer_class=nl.MultiNest)

    # For the final CTI model, constant efficiency mode has a tendency to sample parameter space too fast and infer an
    # inaccurate model. Thus, we turn it off for phase 2.

    phase4.optimizer.const_efficiency_mode = False
    phase4.optimizer.n_live_points = 50
    phase4.optimizer.sampling_efficiency = 0.3

    return pl.Pipeline(pipeline_name, phase1, phase2, phase3, phase4)


from autofit.tools import path_util
from autocti.charge_injection import ci_frame
from autocti.charge_injection import ci_data
from autocti.charge_injection import ci_pattern
from autocti.data import util

from autocti.model import arctic_settings
from autocti.model import arctic_params

import os

# This tool allows one to make simulated charge injection imaging data-sets for calibrating parallel charge transfer
# inefficiency, which can be used to test example pipelines and investigate CTI modeling on data-sets where the
# 'true' answer is known.

# The 'ci_data_type', 'ci_data_model' and 'ci_data_resolution' determine the directory the output data folder, e.g:

# The image will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_image.fits'.
# The noise-map will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_noise_map.fits'.
# The pre cti ci image will be output as '/workspace/data/ci_data_type/ci_data_model/ci_data_resolution/ci_pre_cti.fits'.

# Lets setup the relative path to the workspace, so we can output the data in the 'data' folder..
workspace_path = '{}/../../../../'.format(os.path.dirname(os.path.realpath(__file__)))

ci_data_type = 'ci_images_non_uniform'
ci_data_model = 'parallel_x2_species'
ci_data_resolution = 'high_res'

# Create the path where the data will be output, which in this case is
# '/workspace/data/ci_images_uniform/parallel_x2_species/high_res/'
ci_data_path = path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=['data', ci_data_type, ci_data_model, ci_data_resolution])

if ci_data_resolution is 'high_res':
    shape = (2316, 2119)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 2099)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 2099, 2119)))
elif ci_data_resolution is 'mid_res':
    shape = (2316, 1034)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 1014)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 1014, 1034)))
elif ci_data_resolution is 'low_res':
    shape = (2316, 517)
    frame_geometry = ci_frame.FrameGeometry(corner=(0.0, 0.0),
                                            parallel_overscan=ci_frame.Region((2296, 2316, 51, 497)),
                                            serial_prescan=ci_frame.Region((0, 2316, 0, 51)),
                                            serial_overscan=ci_frame.Region((0, 2296, 497, 517)))

# Specify the charge injection regions on the CCD, which in this case is 7 equally spaced rectangular blocks.
ci_regions = [(0, 30, 51, shape[1]-20), (330, 360, 51, shape[1]-20),
              (660, 690, 51, shape[1]-20), (990, 1020, 51, shape[1]-20),
              (1320, 1350, 51, shape[1]-20), (1650, 1680, 51, shape[1]-20),
              (1980, 2010, 51, shape[1]-20)]

# The normalization of every ci image - this size of this list thus determines how many images are simulated.
normalizations=[100.0, 500.0, 1000.0, 5000.0, 10000.0, 25000.0, 50000.0, 84700.0]

# The CTI settings of arCTIc, which models the CCD read-out including CTI. For parallel ci data, we include 'charge
# injection mode' which accounts for the fact that every pixel is transferred over the full CCD.
parallel_cti_settings = arctic_settings.Settings(well_depth=84700, niter=1, express=2, n_levels=2000,
                                                 charge_injection_mode=True, readout_offset=0)
cti_settings = arctic_settings.ArcticSettings(parallel=parallel_cti_settings)

# The CTI model parameters of arCTIc, which includes each trap species density / lifetime and the CCD properties for
# parallel charge transfer.
parallel_species_0 = arctic_params.Species(trap_density=0.13, trap_lifetime=1.25)
parallel_species_1 = arctic_params.Species(trap_density=0.25, trap_lifetime=4.4)
parallel_ccd = arctic_params.CCD(well_notch_depth=1.0e-4, well_fill_beta=0.58, well_fill_alpha=1.0, well_fill_gamma=0.0)
cti_params = arctic_params.ArcticParams(parallel_species=[parallel_species_0, parallel_species_1],
                                        parallel_ccd=parallel_ccd)

# This function creates uniform blocks of non-charge injection lines, given an input normalization, regions and
# parameters describing its non-uniform properties.
column_deviations = [100.0]*len(normalizations)
row_slopes = [0.0]*len(normalizations)
ci_patterns = ci_pattern.non_uniform_from_lists(normalizations=normalizations, regions=ci_regions, row_slopes=row_slopes)

# Use the simulate ci patterns to generate the pre-cti charge injection images.
ci_pre_ctis = list(map(lambda ci_pattern, column_deviation :
                       ci_pattern.simulate_ci_pre_cti(shape=shape,
                                                      column_deviation=column_deviation,
                                                      maximum_normalization=cti_settings.parallel.well_depth),
                       ci_patterns, column_deviations))

# Use every ci pattern to simulate a ci image.
ci_datas = list(map(lambda ci_pre_cti, ci_pattern:
                    ci_data.simulate(ci_pre_cti=ci_pre_cti, frame_geometry=frame_geometry,
                                     ci_pattern=ci_pattern, cti_settings=cti_settings, cti_params=cti_params,
                                     read_noise=4.0),
                    ci_pre_ctis, ci_patterns))

# Now, output every image to the data folder as the filename 'ci_data_normalization.fits'
list(map(lambda ci_data, normalization :
         util.numpy_array_2d_to_fits(array_2d=ci_data.image,
                                     file_path=ci_data_path + 'image_' + str(int(normalization)) + '.fits',
                                     overwrite=True),
         ci_datas, normalizations))

# Output every pre-cti image to the data folder as the filename 'ci_pre_cti_normalization.fits'. This allows the
# calibration pipeline to load these images as the model pre-cti images, which is necessary for non-uniform ci patterns.
list(map(lambda ci_data, normalization :
         util.numpy_array_2d_to_fits(array_2d=ci_data.ci_pre_cti,
                                     file_path=ci_data_path + 'ci_pre_cti_' + str(int(normalization)) + '.fits',
                                     overwrite=True),
         ci_datas, normalizations))